"""
Contains patches that modify the binary at the data level.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..components.allocation_managers.allocation_manager import MemoryFlag
from .patch import Patch
from .raw_patches import ModifyRawBytesPatch

if TYPE_CHECKING:
    from ..patcherex import Patcherex


class ModifyDataPatch(ModifyRawBytesPatch):
    """
    Extends ModifyRawBytesPatch to only be used for memory addresses, used for modifying data in binary.
    """

    def __init__(self, addr: int, new_bytes: bytes) -> None:
        """
        Same as ModifyRawBytesPatch constructor, but address type of memory is assumed.
        """
        super().__init__(addr, new_bytes, addr_type="mem")


class InsertDataPatch(Patch):
    """
    Patch that inserts data into the binary.
    """

    def __init__(self, addr_or_name: int | str, data: bytes) -> None:
        """
        Constructor.

        :param addr_or_name: If an integer, data is inserted at the address.
                             If a string, it is placed in a free spot in the binary and added as a symbol (with this as its name).
        :param data: New data to place in binary.
        """
        self.addr = None
        self.name = None
        if isinstance(addr_or_name, int):
            self.addr = addr_or_name
        elif isinstance(addr_or_name, str):
            self.name = addr_or_name
        self.data = data

    def apply(self, p: Patcherex) -> None:
        """
        Applies the patch to the binary, intended to be called by a Patcherex instance.

        :param p: Patcherex instance.
        """
        if self.addr:
            p.binfmt_tool.update_binary_content(self.addr, self.data)
        elif self.name:
            block = p.allocation_manager.allocate(len(self.data), flag=MemoryFlag.RW)
            p.symbols[self.name] = block.mem_addr
            p.binfmt_tool.update_binary_content(block.file_addr, self.data)


class RemoveDataPatch(ModifyRawBytesPatch):
    """
    Extends ModifyRawBytesPatch for removing data in the binary (fills it with null bytes starting at address given).
    Expects a memory address.
    """

    def __init__(self, addr: int, size: int) -> None:
        """
        Same as ModifyRawBytes Patch constructor, but adds size parameter and assumes memory address.

        :param size: The number of bytes to remove.
        """
        super().__init__(addr, b"\x00" * size, addr_type="mem")



from __future__ import annotations

import io
import logging

import intelhex

from .binfmt_tool import BinFmtTool

logger = logging.getLogger(__name__)


class IHex(BinFmtTool):
    def __init__(self, p, binary_path: str) -> None:
        super().__init__(p, binary_path)
        self._file = open(binary_path, "rb")
        self._ihex = intelhex.IntelHex(binary_path)
        self.file_size = self._ihex.maxaddr() + 1
        self.file_updates = []

    def __del__(self) -> None:
        self._file.close()

    def _init_memory_analysis(self) -> None:
        pass

    def finalize(self) -> None:
        pass

    def save_binary(self, filename: str | None = None) -> None:
        for update in self.file_updates:
            self._ihex.puts(update["offset"], update["content"])
        if filename is None:
            filename = f"{self.binary_path}.patched"
        sio = io.StringIO()
        self._ihex.write_hex_file(sio, byte_count=0x20)
        final_content = sio.getvalue()
        sio.close()
        entry_point = ""
        final = ""
        for line in final_content.splitlines():
            if line.startswith(":04000005"):
                entry_point = line
            elif line == ":00000001FF":
                final += entry_point + "\n"
                final += line + "\n"
            else:
                final += line + "\n"

        with open(filename, "w") as f:
            f.write(final)

    def update_binary_content(self, offset: int, new_content: bytes) -> None:
        logger.debug(
            f"Updating offset {hex(offset)} with content ({len(new_content)} bytes) {new_content.hex()}"
        )
        for update in self.file_updates:
            if offset >= update["offset"] and offset < update["offset"] + len(
                update["content"]
            ):
                raise ValueError(
                    f"Cannot update offset {hex(offset)} with content {new_content}, it overlaps with a previous update"
                )
        self.file_updates.append({"offset": offset, "content": new_content})
        if offset + len(new_content) > self.file_size:
            self.file_size = offset + len(new_content)

    def get_binary_content(self, offset: int, size: int) -> bytes:
        # check if it's in the file updates
        for update in self.file_updates:
            if offset >= update["offset"] and offset + size <= update["offset"] + len(
                update["content"]
            ):
                return update["content"][
                    offset - update["offset"] : offset - update["offset"] + size
                ]
        # otherwise read from the file
        return self._ihex.tobinarray(start=offset, size=size)

    def append_to_binary_content(self, new_content: bytes) -> None:
        self.file_updates.append({"offset": self.file_size, "content": new_content})
        self.file_size += len(new_content)



class PpcVleInfo:
    nop_bytes = b"\x01\x00\x00\x00"
    nop_size = 4
    jmp_asm = "b {dst}"
    jmp_size = 4
    alignment = 4
    bits = 32
    is_variable_length_isa = True
    instr_size = -1  # variable length
    call_asm = "bl {dst}"
    pc_reg_names = []
    save_context_asm = ""  # TODO
    restore_context_asm = ""  # TODO



class Target:
    target_classes = []

    def __init__(self, p, binary_path):
        self.binary_path = binary_path
        self.p = p

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.target_classes.append(cls)

    @classmethod
    def detect_target(cls, p, binary_path):
        for target_class in cls.target_classes:
            if target_class.detect_target(binary_path):
                return target_class(p, binary_path)
        raise ValueError("Unknown target")

    def get_component(self, component_type, component_name, component_opts=None):
        if component_opts is None:
            component_opts = {}
        return getattr(self, f"get_{component_type}")(component_name, **component_opts)



from __future__ import annotations

import logging
import os
import re
import subprocess
import tempfile

from ..assets.assets import Assets
from .disassembler import Disassembler

logger = logging.getLogger(__name__)


class PpcVle(Disassembler):
    def __init__(self, p) -> None:
        self.p = p
        self.assets_path = Assets("ppc_vle").path

    def disassemble(self, input: bytes, base=0, **kwargs) -> list[dict[str, str | int]]:
        if isinstance(input, str):
            input = bytes(map(ord, input))
        with tempfile.TemporaryDirectory() as td:
            with open(os.path.join(td, "code.bin"), "wb") as f:
                f.write(input)

            try:
                proc = subprocess.run(
                    f"{os.path.join(self.assets_path, 'powerpc-eabivle-objdump')} -D -b binary --adjust-vma={hex(base)} -m powerpc:common -EB {os.path.join(td, 'code')}.bin | tail +8",
                    shell=True,
                    check=True,
                    capture_output=True,
                )
                str_result = proc.stdout.decode("utf-8")
            except subprocess.CalledProcessError as e:
                logger.error(e.stderr.decode("utf-8"))
                raise e

        result = []
        for line in str_result.splitlines():
            m = re.match(
                r"\s+(?P<address>[0-9a-f]+):\s+(?P<bytes>([0-9a-f]{2}\s)+)\s+(?P<mnemonic>.+?)\s+(?P<op_str>.+?)$",
                line,
            )
            if m:
                instr = m.groupdict()
                result.append(
                    {
                        "address": int(instr["address"], 16),
                        "size": len(bytes.fromhex(instr["bytes"])),
                        "mnemonic": re.sub(r"\s+", "", instr["mnemonic"]),
                        "op_str": re.sub(
                            r"\s+", "", instr["op_str"].split(";")[0]
                        ).replace(",", ", "),
                    }
                )
        return result



from __future__ import annotations

import logging

from .compiler import Compiler

logger = logging.getLogger(__name__)


class Clang(Compiler):
    def __init__(
        self, p, clang_version=15, compiler_flags: list[str] | None = None
    ) -> None:
        super().__init__(p)
        self.preserve_none = clang_version >= 19
        if compiler_flags is None:
            compiler_flags = []
        self._compiler = f"clang-{clang_version}"
        self._linker = f"ld.lld-{clang_version}"
        self._compiler_flags = compiler_flags



from .bin_arm_bare import BinArmBare
from .elf_aarch64_linux import ElfAArch64Linux
from .elf_amd64_linux import ElfAmd64Linux
from .elf_amd64_linux_recomp import ElfAmd64LinuxRecomp
from .elf_arm_bare import ElfArmBare
from .elf_arm_linux import ElfArmLinux
from .elf_arm_linux_recomp import ElfArmLinuxRecomp
from .elf_arm_mimxrt1052 import ElfArmMimxrt1052
from .elf_leon3_bare import ElfLeon3Bare
from .elf_mips64_linux import ElfMips64Linux
from .elf_mips64el_linux import ElfMips64elLinux
from .elf_mips_linux import ElfMipsLinux
from .elf_mipsel_linux import ElfMipselLinux
from .elf_ppc64_linux import ElfPpc64Linux
from .elf_ppc64le_linux import ElfPpc64leLinux
from .elf_ppc_linux import ElfPpcLinux
from .elf_x86_linux import ElfX86Linux
from .ihex_ppc_bare import IHexPPCBare
from .target import Target

__all__ = [
    "BinArmBare",
    "ElfAArch64Linux",
    "ElfAmd64Linux",
    "ElfAmd64LinuxRecomp",
    "ElfArmBare",
    "ElfArmLinux",
    "ElfArmLinuxRecomp",
    "ElfArmMimxrt1052",
    "ElfLeon3Bare",
    "ElfMips64Linux",
    "ElfMips64elLinux",
    "ElfMipsLinux",
    "ElfMipselLinux",
    "ElfPpc64Linux",
    "ElfPpc64leLinux",
    "ElfPpcLinux",
    "ElfX86Linux",
    "IHexPPCBare",
    "Target",
]



#!/usr/bin/env python

# ruff: noqa
import logging
import os
import shutil
import subprocess
import tempfile

import pytest

from patcherex2 import *

logging.getLogger("patcherex2").setLevel("DEBUG")


class Tests:
    @pytest.fixture(autouse=True, scope="class", params=["angr", "ghidra"])
    def setup(self, request):
        request.cls.bin_location = str(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "./test_binaries/mips",
            )
        )
        request.cls.binary_analyzer = request.param

    def test_raw_file_patch_nopie(self):
        self.run_one(
            "printf_nopie",
            [ModifyRawBytesPatch(0x81B, b"No", addr_type="raw")],
            expected_output=b"No",
            expected_returnCode=0,
        )

    def test_raw_file_patch_pie(self):
        self.run_one(
            "printf_pie",
            [ModifyRawBytesPatch(0x8EF, b"No", addr_type="raw")],
            expected_output=b"No",
            expected_returnCode=0,
        )

    def test_raw_mem_patch_nopie(self):
        self.run_one(
            "printf_nopie",
            [ModifyRawBytesPatch(0x40081B, b"No")],
            expected_output=b"No",
            expected_returnCode=0,
        )

    def test_raw_mem_patch_pie(self):
        self.run_one(
            "printf_pie",
            [ModifyRawBytesPatch(0x8EF, b"No")],
            expected_output=b"No",
            expected_returnCode=0,
        )

    def test_modify_instruction_patch_nopie(self):
        self.run_one(
            "printf_nopie",
            [
                ModifyInstructionPatch(0x400710, "addiu $a1, $at, 0x818"),
            ],
            expected_output=b"%s",
            expected_returnCode=0,
        )

    def test_modify_instruction_patch_pie(self):
        self.run_one(
            "printf_pie",
            [
                ModifyInstructionPatch(0x7E0, "addiu $a1, $at, 0x8EC"),
            ],
            expected_output=b"%s",
            expected_returnCode=0,
        )

    def test_insert_instruction_patch_nopie(self):
        instrs = """
            li $v0, 0xfa4
            li $a0, 0x1
            la $a1, 0x40081b
            li $a2, 0x3
            syscall
        """
        self.run_one(
            "printf_nopie",
            [InsertInstructionPatch(0x400720, instrs)],
            expected_output=b"Hi\x00Hi",
            expected_returnCode=0,
        )

    def test_insert_instruction_patch_pie(self):
        instrs = """
            li $v0, 0xfa4
            li $a0, 0x1
            li $a1, 0x0
            move $a1, $s4
            li $a3, 0xfff
            not $a3, $a3
            and $a1, $a1, $a3
            addiu $a1, $a1, 0x8EF
            li $a2, 0x3
            syscall
        """
        self.run_one(
            "printf_pie",
            [InsertInstructionPatch(0x7F0, instrs)],
            expected_output=b"Hi\x00Hi",
            expected_returnCode=0,
        )

    def test_insert_instruction_patch_2_nopie(self):
        instrs = """
            li $v0, 0xfa1
            li $a0, 0x32
            syscall
        """
        self.run_one(
            "printf_nopie",
            [
                InsertInstructionPatch("return_0x32", instrs),
                ModifyInstructionPatch(0x400720, "j {return_0x32}"),
            ],
            expected_returnCode=0x32,
        )

    def test_insert_instruction_patch_2_pie(self):
        instrs = """
            li $v0, 0xfa1
            li $a0, 0x32
            syscall
        """
        self.run_one(
            "printf_pie",
            [
                InsertInstructionPatch("return_0x32", instrs),
                ModifyInstructionPatch(0x7F0, "j {return_0x32}"),
            ],
            expected_returnCode=0x32,
        )

    def test_remove_instruction_patch_nopie(self):
        self.run_one(
            "printf_nopie",
            [
                RemoveInstructionPatch(0x40081C, num_bytes=4),
            ],
            expected_output=b"H",
            expected_returnCode=0,
        )

    def test_remove_instruction_patch_pie(self):
        self.run_one(
            "printf_pie",
            [
                RemoveInstructionPatch(0x8F0, num_bytes=4),
            ],
            expected_output=b"H",
            expected_returnCode=0,
        )

    def test_modify_data_patch_nopie(self):
        self.run_one(
            "printf_nopie",
            [ModifyDataPatch(0x40081B, b"No")],
            expected_output=b"No",
            expected_returnCode=0,
        )

    def test_modify_data_patch_pie(self):
        self.run_one(
            "printf_pie",
            [ModifyDataPatch(0x8EF, b"No")],
            expected_output=b"No",
            expected_returnCode=0,
        )

    def test_insert_data_patch_nopie(self, tlen=5):
        p1 = InsertDataPatch("added_data", b"A" * tlen)
        instrs = """
            li $v0, 0xfa4
            li $a0, 0x1
            li $a1, {added_data}
            li $a2, %s
            syscall
        """ % hex(tlen)
        p2 = InsertInstructionPatch(0x400720, instrs)
        self.run_one(
            "printf_nopie",
            [p1, p2],
            expected_output=b"A" * tlen + b"Hi",
            expected_returnCode=0,
        )

    @pytest.mark.skip(reason="difficult to do with pie")
    def test_insert_data_patch_pie(self, tlen=5):
        p1 = InsertDataPatch("added_data", b"A" * tlen)
        instrs = """
            li $v0, 0xfa4
            li $a0, 0x1
            li $a1, {added_data}
            li $a2, %s
            syscall
        """ % hex(tlen)
        p2 = InsertInstructionPatch(0x7F0, instrs)
        self.run_one(
            "printf_pie",
            [p1, p2],
            expected_output=b"A" * tlen + b"Hi",
            expected_returnCode=0,
        )

    def test_remove_data_patch_nopie(self):
        self.run_one(
            "printf_nopie",
            [RemoveDataPatch(0x40081C, 1)],
            expected_output=b"H",
            expected_returnCode=0,
        )

    def test_remove_data_patch_pie(self):
        self.run_one(
            "printf_pie",
            [RemoveDataPatch(0x8F0, 1)],
            expected_output=b"H",
            expected_returnCode=0,
        )

    def test_replace_function_patch(self):
        code = """
        int add(int a, int b){ for(;; b--, a+=2) if(b <= 0) return a; }
        """
        self.run_one(
            "replace_function_patch",
            [ModifyFunctionPatch(0x860, code)],
            expected_output=b"70707070",
            expected_returnCode=0,
        )

    @pytest.mark.skip(reason="waiting for cle relocation support")
    def test_replace_function_patch_with_function_reference(self):
        code = """
        extern int add(int, int);
        extern int subtract(int, int);
        int multiply(int a, int b){ for(int c = 0;; b = subtract(b, 1), c = subtract(c, a)) if(b <= 0) return c; }
        """
        self.run_one(
            "replace_function_patch",
            [ModifyFunctionPatch(0x970, code)],
            expected_output=b"-21-21",
            expected_returnCode=0,
        )

    @pytest.mark.skip(reason="waiting for cle relocation support")
    def test_replace_function_patch_with_function_reference_and_rodata(self):
        code = """
        extern int printf(const char *format, ...);
        int multiply(int a, int b){ printf("%sWorld %s %s %s %d\\n", "Hello ", "Hello ", "Hello ", "Hello ", a * b);printf("%sWorld\\n", "Hello "); return a * b; }
        """
        self.run_one(
            "replace_function_patch",
            [ModifyFunctionPatch(0x970, code)],
            expected_output=b"Hello World Hello  Hello  Hello  21\nHello World\n2121",
            expected_returnCode=0,
        )

    @pytest.mark.skip(reason="waiting for cle relocation support")
    def test_insert_function_patch(self):
        insert_code = """
        int min(int a, int b) { return (a < b) ? a : b; }
        """
        replace_code = """
        extern int min(int, int);
        int max(int a, int b) { return min(a, b); }
        """
        self.run_one(
            "replace_function_patch",
            [
                InsertFunctionPatch("min", insert_code),
                ModifyFunctionPatch(0xAD4, replace_code),
            ],
            expected_output=b"2121212121",
            expected_returnCode=0,
        )

    def run_one(
        self,
        filename,
        patches,
        set_oep=None,
        inputvalue=None,
        expected_output=None,
        expected_returnCode=None,
    ):
        filepath = os.path.join(self.bin_location, filename)
        pipe = subprocess.PIPE

        with tempfile.TemporaryDirectory() as td:
            tmp_file = os.path.join(td, "patched")
            p = Patcherex(
                filepath, target_opts={"binary_analyzer": self.binary_analyzer}
            )
            for patch in patches:
                p.patches.append(patch)
            p.apply_patches()
            p.save_binary(tmp_file)
            p.shutdown()
            # os.system(f"readelf -hlS {tmp_file}")

            p = subprocess.Popen(
                ["qemu-mips", "-L", "/usr/mips-linux-gnu", tmp_file],
                stdin=pipe,
                stdout=pipe,
                stderr=pipe,
            )
            res = p.communicate(inputvalue)
            if expected_output:
                if res[0] != expected_output:
                    pytest.fail(
                        f"AssertionError: {res[0]} != {expected_output}, binary dumped: {self.dump_file(tmp_file)}"
                    )
                # self.assertEqual(res[0], expected_output)
            if expected_returnCode:
                if p.returncode != expected_returnCode:
                    pytest.fail(
                        f"AssertionError: {p.returncode} != {expected_returnCode}, binary dumped: {self.dump_file(tmp_file)}"
                    )
                # self.assertEqual(p.returncode, expected_returnCode)

    def dump_file(self, file):
        shutil.copy(file, "/tmp/patcherex_failed_binary")
        return "/tmp/patcherex_failed_binary"



from __future__ import annotations

import enum
import logging
from pprint import pformat

logger = logging.getLogger(__name__)


class Block:
    subclasses = []

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        Block.subclasses.append(cls)

    def __init__(self, addr: int, size: int, is_free=True) -> None:
        self.addr = addr
        self.size = size
        self.is_free = is_free

    def __lt__(self, other: Block) -> bool:
        return self.addr < other.addr

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} addr={hex(self.addr)} size={hex(self.size)} is_free={self.is_free}>"

    def coalesce(self, other: Block) -> bool:
        if self.is_free == other.is_free and self.addr + self.size == other.addr:
            self.size += other.size
            return True
        return False


class FileBlock(Block):
    pass


class MemoryBlock(Block):
    def __init__(self, addr: int, size: int, is_free=True) -> None:
        super().__init__(addr, size, is_free)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} addr={hex(self.addr)} size={hex(self.size)} is_free={self.is_free}>"


class MemoryFlag(enum.IntFlag):
    UNDEF = enum.auto()
    R = 0x4
    W = 0x2
    X = 0x1
    RW = R | W
    RX = R | X
    RWX = R | W | X


class MappedBlock(Block):
    def __init__(
        self,
        file_addr: int,
        mem_addr: int,
        size: int,
        is_free=True,
        flag=None,
        load_mem_addr: int = None,
    ) -> None:
        """
        :param file_addr: file address of the block
        :param mem_addr: virtual memory address of the block
        :param size: size of the block
        :param is_free: whether the block is free
        :param flag: memory flag of the block
        :param load_mem_addr: load memory address of the block, if not provided, it will be the same as mem_addr
        """
        super().__init__(None, size, is_free)
        self.file_addr = file_addr
        self.mem_addr = mem_addr
        self.flag = flag
        self.load_mem_addr = load_mem_addr if load_mem_addr else mem_addr

    def __lt__(self, other: MappedBlock) -> bool:
        return self.mem_addr < other.mem_addr

    def __repr__(self) -> str:
        repr = f"<{self.__class__.__name__} file_addr={hex(self.file_addr)} mem_addr={hex(self.mem_addr)} size={hex(self.size)} is_free={self.is_free} flag={str(self.flag)}"
        if self.load_mem_addr != self.mem_addr:
            repr += f" load_mem_addr={hex(self.load_mem_addr)}"
        repr += ">"
        return repr

    def coalesce(self, other: MappedBlock) -> bool:
        if (
            self.flag == other.flag
            and self.is_free == other.is_free
            and self.file_addr + self.size == other.file_addr
            and self.mem_addr + self.size == other.mem_addr
        ):
            self.size += other.size
            return True
        return False


class AllocationManager:
    def __init__(self, p) -> None:
        self.blocks = {cls: [] for cls in Block.subclasses}
        self.p = p
        self.new_mapped_blocks = []

    def add_block(self, block: Block) -> None:
        self.blocks[type(block)].append(block)
        self.blocks[type(block)].sort()
        self.coalesce(self.blocks[type(block)])

    def add_free_space(self, addr: int, size: int, flag: str = "RX") -> None:
        _flag = 0
        if "r" in flag.lower():
            _flag |= MemoryFlag.R
        if "w" in flag.lower():
            _flag |= MemoryFlag.W
        if "x" in flag.lower():
            _flag |= MemoryFlag.X
        block = MappedBlock(
            self.p.binary_analyzer.mem_addr_to_file_offset(addr),
            addr,
            size,
            is_free=True,
            flag=_flag,
        )
        self.p.allocation_manager.add_block(block)

    def _find_in_mapped_blocks(
        self, size: int, flag=MemoryFlag.RWX, align=0x1
    ) -> MappedBlock:
        best_fit = None
        for block in self.blocks[MappedBlock]:
            if block.is_free and block.size >= size and block.flag & flag == flag:
                # check for alignment
                offset = (align - (block.mem_addr % align)) % align
                if block.size >= size + offset:
                    if block.size == size + offset and offset > 0:
                        block.is_free = False
                        return block
                    elif best_fit is None or block.size < best_fit.size:
                        best_fit = block

        if best_fit:
            # Adjust for alignment
            offset = (align - (best_fit.mem_addr % align)) % align
            remaining_size = best_fit.size - size - offset
            allocated_block = MappedBlock(
                best_fit.file_addr + offset,
                best_fit.mem_addr + offset,
                size,
                is_free=False,
                flag=flag,
            )
            self.add_block(allocated_block)
            if offset > 0:
                self.add_block(
                    MappedBlock(
                        best_fit.file_addr,
                        best_fit.mem_addr,
                        offset,
                        is_free=True,
                        flag=flag,
                    )
                )
            best_fit.file_addr += size + offset
            best_fit.mem_addr += size + offset
            best_fit.size = remaining_size
            if best_fit.size == 0:
                self.blocks[MappedBlock].remove(best_fit)
            return allocated_block

    def _create_new_mapped_block(
        self, size: int, flag=MemoryFlag.RWX, align=0x1
    ) -> bool:
        # TODO: currently we won't use available file/mem blocks, instead we create new one at the end of the file
        file_addr = None
        mem_addr = None
        for block in self.blocks[FileBlock]:
            if block.size == -1:
                file_addr = block.addr
                block.addr += 0x10000
        for block in self.blocks[MemoryBlock]:
            if block.size == -1:
                # NOTE: mem_addr % p_align should equal to file_addr % p_align
                # Check `man elf` and search for `p_align` for more information
                # FIXME: shouldn't do any assumption on component type, reimpl in a better way
                # FIXME: even worse, importing ELF will cause circular import
                # TODO: consider merge allocation_manager and binfmt_tool into one component
                if self.p.binfmt_tool.__class__.__name__ == "ELF":
                    max_seg_align = max(
                        [segment["p_align"] for segment in self.p.binfmt_tool._segments]
                        + [0]
                    )
                    mem_addr = block.addr + (file_addr - block.addr) % max_seg_align
                else:
                    mem_addr = block.addr + (file_addr - block.addr) % 0x1000
                block.addr = mem_addr + 0x10000
        if file_addr and mem_addr:
            self.add_block(
                MappedBlock(file_addr, mem_addr, 0x10000, is_free=True, flag=flag)
            )
            self.new_mapped_blocks.append(
                MappedBlock(file_addr, mem_addr, 0x10000, is_free=True, flag=flag)
            )
            return True
        return False

    def allocate(self, size: int, flag=MemoryFlag.RWX, align=0x1) -> MappedBlock:
        logger.debug(
            f"allocating size: {size}, flag: {flag.__repr__()}, align: {align}"
        )
        block = self._find_in_mapped_blocks(size, flag, align)
        if block:
            return block
        logger.debug(
            f"memory_allocate: failed to allocate memory of size {size} with flag {flag.__repr__()}, creating new area and retrying"
        )
        if self._create_new_mapped_block(size, flag, align):
            return self.allocate(size, flag, align)
        else:
            raise MemoryError("Insufficient memory")

    def free(self, block: Block) -> None:
        block.is_free = True
        self.coalesce(self.blocks[type(block)])

    def coalesce(self, blocks: list[Block]) -> None:
        for curr, next in zip(blocks, blocks[1:]):
            if curr.coalesce(next):
                blocks.remove(next)
                self.coalesce(blocks)
                return

    def finalize(self) -> None:
        for block in self.new_mapped_blocks:
            for mapped_block in self.blocks[MappedBlock]:
                if mapped_block.is_free:
                    if (
                        block.mem_addr + block.size
                        == mapped_block.mem_addr + mapped_block.size
                        and block.mem_addr <= mapped_block.mem_addr
                    ):
                        self.blocks[MappedBlock].remove(mapped_block)
                        block.size -= mapped_block.size
                        return self.finalize()

        for block in self.new_mapped_blocks:
            if block.file_addr + block.size > self.p.binfmt_tool.file_size:
                self.p.binfmt_tool.file_size = block.file_addr + block.size

        logger.debug(f"finalized blocks: \n{pformat(list(self.blocks.values()))}")
        logger.debug(f"new mapped blocks: \n{pformat(self.new_mapped_blocks)}")



class Ppc64Info:
    nop_bytes = b"\x60\x00\x00\x00"
    nop_size = 4
    jmp_asm = "b {dst}"
    jmp_size = 4
    alignment = 4
    bits = 64
    is_variable_length_isa = False
    instr_size = 4
    call_asm = "bl {dst}"
    pc_reg_names = []
    save_context_asm = """
    stwu r1, -0x80(r1)
    stmw r3, 0x8(r1)
    """
    restore_context_asm = """
    lmw r3, 0x8(r1)
    addi r1, r1, 0x80
    """



"""
Contains patches that modify the binary at the function level.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ..components.allocation_managers.allocation_manager import MemoryFlag
from .patch import Patch

if TYPE_CHECKING:
    from ..patcherex import Patcherex

logger = logging.getLogger(__name__)


class ModifyFunctionPatch(Patch):
    """
    Patch that replaces an existing function in the binary with your own. If there is enough room in the existing
    function, your code is compiled and placed there. If not, your code is placed in a free spot in the binary, and
    the function will jump there instead.
    """

    def __init__(
        self,
        addr_or_name: int | str,
        code: str,
        detour_pos=-1,
        symbols: dict[str, int] | None = None,
        **kwargs,
    ) -> None:
        """
        Constructor.

        :param addr_or_name: The name or file address of the function.
        :param code: C code to replace the function.
        :param detour_pos: If original function is not big enough, file address to place the given code, defaults to -1
        :param symbols: Symbols to include when compiling, in format {symbol name: memory address}, defaults to None
        """
        self.code = code
        self.detour_pos = detour_pos
        self.addr_or_name = addr_or_name
        self.symbols = symbols if symbols else {}
        self.compile_opts = kwargs["compile_opts"] if "compile_opts" in kwargs else {}

    def apply(self, p: Patcherex) -> None:
        """
        Applies the patch to the binary, intended to be called by a Patcherex instance.

        :param p: Patcherex instance.
        """
        func = p.binary_analyzer.get_function(self.addr_or_name)
        compiled_size = len(
            p.compiler.compile(
                self.code,
                symbols=self.symbols,
                is_thumb=p.binary_analyzer.is_thumb(func["addr"]),
                **self.compile_opts,
            )
        )
        if compiled_size <= func["size"]:
            mem_addr = func["addr"]
            file_addr = p.binary_analyzer.mem_addr_to_file_offset(mem_addr)
        else:
            # TODO: mark the function as free (exclude jump instr)
            if self.detour_pos == -1:
                block = p.allocation_manager.allocate(
                    compiled_size + 0x20, align=0x4, flag=MemoryFlag.RX
                )
                mem_addr = block.mem_addr
                file_addr = block.file_addr
            else:
                mem_addr = self.detour_pos
                file_addr = p.binary_analyzer.mem_addr_to_file_offset(mem_addr)
            jmp_instr = p.archinfo.jmp_asm.format(dst=hex(mem_addr))
            jmp_bytes = p.assembler.assemble(
                jmp_instr,
                func["addr"],
                is_thumb=p.binary_analyzer.is_thumb(func["addr"]),
            )
            p.binfmt_tool.update_binary_content(
                p.binary_analyzer.mem_addr_to_file_offset(func["addr"]),
                jmp_bytes,
            )
        p.binfmt_tool.update_binary_content(
            file_addr,
            p.compiler.compile(
                self.code,
                mem_addr,
                symbols=self.symbols,
                is_thumb=p.binary_analyzer.is_thumb(func["addr"]),
                **self.compile_opts,
            ),
        )


class InsertFunctionPatch(Patch):
    """
    Inserts a function into the binary.
    """

    def __init__(
        self,
        addr_or_name: int | str,
        code: str,
        force_insert=False,
        detour_pos=-1,
        symbols: dict[str, int] | None = None,
        is_thumb=False,
        **kwargs,
    ) -> None:
        """
        Constructor.

        :param addr_or_name: If an integer, an intermediate function is created in a free spot in the binary,
                             and at that address, a jump to the function is made with necessary context saves.
                             If a string, the function is created in a free spot in the binary with that name.
        :param code: C code for the new function. "SAVE_CONTEXT" and "RESTORE_CONTEXT" can be used to save and restore context.
        :param force_insert: If Patcherex should ignore whether instructions can be moved when inserting, defaults to False
        :param detour_pos: If address is used, this is the address to place trampoline code for jumping to function.
                           If name is used, this is where the new function will be placed, defaults to -1
        :param symbols: Symbols to include when compiling/assembling, in format {symbol name: memory address}, defaults to None
        :param is_thumb: Whether the instructions given are thumb, defaults to False
        :param kwargs: Extra options. Can include "prefunc" and "postfunc", instructions to go before or after your function if you give an address.
                         Can also have "save_context" for whether context should be saved and "compile_opts" for extra compile options.
        """
        self.addr = None
        self.name = None
        if isinstance(addr_or_name, int):
            self.addr = addr_or_name
        elif isinstance(addr_or_name, str):
            self.name = addr_or_name
        self.code = code
        self.detour_pos = detour_pos
        self.symbols = symbols if symbols else {}
        self.is_thumb = is_thumb
        self.force_insert = force_insert
        self.prefunc = kwargs["prefunc"] if "prefunc" in kwargs else None
        self.postfunc = kwargs["postfunc"] if "postfunc" in kwargs else None
        self.compile_opts = kwargs["compile_opts"] if "compile_opts" in kwargs else {}
        self.save_context = (
            kwargs["save_context"] if "save_context" in kwargs else False
        )

    def apply(self, p: Patcherex) -> None:
        """
        Applies the patch to the binary, intended to be called by a Patcherex instance.

        :param p: Patcherex instance.
        """
        if self.addr:
            if self.prefunc:
                if "SAVE_CONTEXT" in self.prefunc:
                    self.prefunc = self.prefunc.replace(
                        "SAVE_CONTEXT", f"\n{p.archinfo.save_context_asm}\n"
                    )
                if "RESTORE_CONTEXT" in self.prefunc:
                    self.prefunc = self.prefunc.replace(
                        "RESTORE_CONTEXT", f"\n{p.archinfo.restore_context_asm}\n"
                    )
            if self.postfunc:
                if "SAVE_CONTEXT" in self.postfunc:
                    self.postfunc = self.postfunc.replace(
                        "SAVE_CONTEXT", f"\n{p.archinfo.save_context_asm}\n"
                    )
                if "RESTORE_CONTEXT" in self.postfunc:
                    self.postfunc = self.postfunc.replace(
                        "RESTORE_CONTEXT", f"\n{p.archinfo.restore_context_asm}\n"
                    )
            ifp = InsertFunctionPatch(
                f"__patcherex_{hex(self.addr)}",
                self.code,
                is_thumb=p.binary_analyzer.is_thumb(self.addr),
                symbols=self.symbols,
            )
            ifp.apply(p)
            instrs = ""
            instrs += p.archinfo.save_context_asm if self.save_context else ""
            instrs += self.prefunc if self.prefunc else ""
            instrs += "\n"
            # NOTE: ↓ this is hardcoded to bl, not blx, but it should be fine for this use case
            instrs += p.archinfo.call_asm.format(
                dst=f"{{__patcherex_{hex(self.addr)}}}"
            )
            instrs += "\n"
            instrs += self.postfunc if self.postfunc else ""
            instrs += p.archinfo.restore_context_asm if self.save_context else ""
            p.utils.insert_trampoline_code(
                self.addr,
                instrs,
                force_insert=self.force_insert,
                detour_pos=self.detour_pos,
                symbols=self.symbols,
            )
        elif self.name:
            compiled_size = len(
                p.compiler.compile(
                    self.code,
                    symbols=self.symbols,
                    is_thumb=self.is_thumb,
                    **self.compile_opts,
                )
            )
            if self.detour_pos == -1:
                block = p.allocation_manager.allocate(
                    compiled_size + 0x20, align=p.archinfo.alignment, flag=MemoryFlag.RX
                )  # TODO: adjust that 0x20 part
                mem_addr = block.mem_addr
                file_addr = block.file_addr
            else:
                mem_addr = self.detour_pos
                file_addr = p.binary_analyzer.mem_addr_to_file_offset(mem_addr)
            p.sypy_info["patcherex_added_functions"].append(hex(mem_addr))
            p.symbols[self.name] = mem_addr
            p.binfmt_tool.update_binary_content(
                file_addr,
                p.compiler.compile(
                    self.code,
                    mem_addr,
                    symbols=self.symbols,
                    is_thumb=self.is_thumb,
                    **self.compile_opts,
                ),
            )


class RemoveFunctionPatch(Patch):
    """
    Patch that removes a function from the binary. Not implemented.
    """

    def __init__(self, parent=None) -> None:
        """
        Constructor.
        """
        raise NotImplementedError()



from patcherex2 import *
import logging

logger = logging.getLogger("patcherex2.patches.instruction_patches")
logger.setLevel(logging.INFO)

p = Patcherex("add", target_opts={"compiler": "clang19"})

c_forward_header = """
// This string will be inserted outside the micropatch function. It will be inserted before your code.
// This is how you can define types and function forward declarations used by your C micropatch
#include <stdio.h>
"""

# The asm_header is inserted in the main body of the patch before the C code. This header is primarily
# useful for gaining access to the stack pointer, which is a register that is unavailable in our C
# code. In this example we have moved rsp to the r12 register, which is a register that is accessible.
# This means that inside the C code we can access variables on the stack by using the r10 variable.
# There is also an asm_footer
asm_header = "mov r12, rsp"

# We can access assembly registers directly by using their name, while still using high level C constructs
# as well as intermediate variables. Note that you can use a return statement anywhere in your C micropatch
# to jump back to the next instruction after the micropatch insertion point.
c_str = """
rdi += rdi;
rdi += 5;
// Print out rsp as it was before the patch was started
printf("%p\\n", (void *) r12);
"""

# It is generally a good idea to mark some registers as scratch to give the compiler
# breathing room for allocating registers to use for intermediate variables in your micropatch
# All of the registers that we mark as scratch can be freely clobbered by the compiler
# Note that you can still read from scratch registers stored in the variables. What the scratch
# register denotation will indicate however is that the register can be re-used after the variable
# is no longer live.
c_scratch_regs = [
    'r8', 'r9', 'r10', 'r11', 'r13', 'r14', 'r15'
    'xmm0', 'xmm1', 'xmm2', 'xmm3', 'xmm4', 'xmm5', 'xmm6', 'xmm7', 'xmm9', 'xmm10', 'xmm11', 'xmm12', 'xmm13', 'xmm14', 'xmm15'
]

# By default floating point registers will have the 'float' type. We can use c_regs_sort to override
# certain registers so they hold different types. In this example we denote that xmm8 is of type double
c_regs_sort = [('xmm8', 'double')]

config = InsertInstructionPatch.CConfig(
    c_forward_header = c_forward_header,
    scratch_regs=c_scratch_regs,
    regs_sort=c_regs_sort,
    asm_header=asm_header
)

p.patches.append(InsertInstructionPatch(0x114d, c_str, language="C", c_config=config))
p.apply_patches()

p.binfmt_tool.save_binary()



class SparcInfo:
    nop_bytes = b"\x01\x00\x00\x00"
    nop_size = 4
    jmp_asm = "b {dst}\nnop"  # nop due to delay slot
    jmp_size = 8
    alignment = 4
    is_variable_length_isa = False
    instr_size = 4
    call_asm = "call {dst}"
    pc_reg_names = ["pc"]
    save_context_asm = ""  # TODO
    restore_context_asm = ""  # TODO



from __future__ import annotations

import logging
import os
import subprocess
import tempfile

import cle
from elftools.elf.elffile import ELFFile

logger = logging.getLogger(__name__)


class Compiler:
    def __init__(self, p) -> None:
        self.p = p
        # preserve_none is a special attribute flag to allow us to control more registers as input to a C function
        # This feature is used for a C instruction patch
        self.preserve_none = False

    def compile(
        self,
        code: str,
        base=0,
        symbols: dict[str, int] | None = None,
        extra_compiler_flags: list[str] | None = None,
        **kwargs,
    ) -> bytes:
        if symbols is None:
            symbols = {}
        if extra_compiler_flags is None:
            extra_compiler_flags = []
        with tempfile.TemporaryDirectory() as td:
            # source file
            with open(os.path.join(td, "code.c"), "w") as f:
                f.write(code)

            # compile to object file
            try:
                args = (
                    [self._compiler]
                    + self._compiler_flags
                    + extra_compiler_flags
                    + [
                        "-c",
                        os.path.join(td, "code.c"),
                        "-o",
                        os.path.join(td, "obj.o"),
                    ]
                )
                subprocess.run(args, check=True, capture_output=True)
            except subprocess.CalledProcessError as e:
                logger.error(e.stderr.decode("utf-8"))
                raise e

            # linker script
            _symbols = {}
            _symbols.update(self.p.symbols)
            _symbols.update(self.p.binary_analyzer.get_all_symbols())
            _symbols.update(symbols)

            # TODO: shouldn't put .rodata in .text, but otherwise switch case jump table won't work
            # Note that even we don't include .rodata here, cle might still include it if there is
            # no gap between .text and .rodata
            with open(os.path.join(td, "obj.o"), "rb") as f:
                linker_script_rodata_sections = " ".join(
                    [
                        f". = ALIGN({section['sh_addralign']}); *({section.name})"
                        for section in ELFFile(f).iter_sections()
                        if section.name.startswith(".rodata")
                    ]
                )
            linker_script_symbols = "".join(
                f"{name} = {hex(addr)};" for name, addr in _symbols.items()
            )

            linker_script = f"SECTIONS {{ .patcherex2 : SUBALIGN(0) {{ . = {hex(base)}; *(.text) {linker_script_rodata_sections} {linker_script_symbols} }} }}"
            with open(os.path.join(td, "linker.ld"), "w") as f:
                f.write(linker_script)

            # link object file
            try:
                args = [self._linker] + [
                    "-relocatable",
                    os.path.join(td, "obj.o"),
                    "-T",
                    os.path.join(td, "linker.ld"),
                    "-o",
                    os.path.join(td, "obj_linked.o"),
                ]
                subprocess.run(args, check=True, capture_output=True)
            except subprocess.CalledProcessError as e:
                logger.error(e.stderr.decode("utf-8"))
                raise e

            # extract compiled code
            ld = cle.Loader(
                os.path.join(td, "obj_linked.o"), main_opts={"base_addr": 0x0}
            )

            patcherex2_section = next(
                (s for s in ld.main_object.sections if s.name == ".patcherex2"), None
            )
            compiled_start = ld.all_objects[0].entry + base

            compiled = ld.memory.load(
                compiled_start,
                patcherex2_section.memsize - compiled_start,
            )
        return compiled



from __future__ import annotations

import logging
import os

from elftools.construct.lib import Container
from elftools.elf.constants import P_FLAGS, SH_FLAGS
from elftools.elf.elffile import ELFFile

from ..allocation_managers.allocation_manager import (
    FileBlock,
    MappedBlock,
    MemoryBlock,
    MemoryFlag,
)
from .binfmt_tool import BinFmtTool

logger = logging.getLogger(__name__)


class ELF(BinFmtTool):
    def __init__(self, p, binary_path: str) -> None:
        super().__init__(p, binary_path)
        self._file = open(binary_path, "rb")
        self._elf = ELFFile(self._file)
        self._segments = [segment.header for segment in self._elf.iter_segments()]
        self._sections = [section.header for section in self._elf.iter_sections()]
        self.file_updates = []

        self.file_size = os.stat(self.binary_path).st_size
        with open(self.binary_path, "rb") as f:
            self.original_binary_content = f.read()
        self.updated_binary_content = self.original_binary_content
        self._init_memory_analysis()

    def __del__(self) -> None:
        self._file.close()

    def _find_space_between_sections(self) -> None:
        load_segments = sorted(
            (
                segment
                for segment in self._elf.iter_segments()
                if segment["p_type"] == "PT_LOAD"
            ),
            key=lambda x: x["p_offset"],
        )

        alloc_sections = sorted(
            (
                section
                for section in self._elf.iter_sections()
                if section["sh_flags"] & SH_FLAGS.SHF_ALLOC
            ),
            key=lambda x: (x["sh_offset"], x["sh_size"]),
        )

        # Spaces in LOAD segments AND between sections
        for segment in load_segments:
            segment_file_end = segment["p_offset"] + segment["p_filesz"]
            segment_mem_end = segment["p_vaddr"] + segment["p_memsz"]
            for curr_sec, next_sec in zip(alloc_sections, alloc_sections[1:]):
                if segment.section_in_segment(curr_sec) and segment.section_in_segment(
                    next_sec
                ):
                    gap_file_start = curr_sec["sh_offset"] + curr_sec["sh_size"]
                    gap_file_size = next_sec["sh_offset"] - gap_file_start
                    gap_mem_start = curr_sec["sh_addr"] + curr_sec["sh_size"]
                    gap_mem_size = next_sec["sh_addr"] - gap_mem_start
                    if (
                        gap_mem_size > 0
                        and gap_file_size > 0
                        and segment_file_end > gap_file_start
                        and segment_mem_end > gap_mem_start
                    ):
                        flag = (
                            MemoryFlag.RW
                            if segment["p_flags"] & P_FLAGS.PF_W
                            else MemoryFlag.RX
                        )
                        block = MappedBlock(
                            gap_file_start,
                            gap_mem_start,
                            min(gap_file_size, gap_mem_size),
                            is_free=True,
                            flag=flag,
                        )
                        self.p.allocation_manager.add_block(block)
            for sec in alloc_sections:
                if segment.section_in_segment(sec):
                    flag = (
                        MemoryFlag.RW
                        if segment["p_flags"] & P_FLAGS.PF_W
                        else MemoryFlag.RX
                    )
                    block = MappedBlock(
                        sec["sh_offset"],
                        sec["sh_addr"],
                        sec["sh_size"],
                        is_free=False,
                        flag=flag,
                    )
                    self.p.allocation_manager.add_block(block)

        # of course also before and after the first and last section in a LOAD segment
        for segment in load_segments:
            segment_file_end = segment["p_offset"] + segment["p_filesz"]
            segment_mem_end = segment["p_vaddr"] + segment["p_memsz"]
            first_sec = next(
                (
                    section
                    for section in alloc_sections
                    if segment.section_in_segment(section)
                ),
                None,
            )
            last_sec = next(
                (
                    section
                    for section in reversed(alloc_sections)
                    if segment.section_in_segment(section)
                ),
                None,
            )
            if first_sec:
                gap_file_start = segment["p_offset"]
                gap_file_size = first_sec["sh_offset"] - gap_file_start
                gap_mem_start = segment["p_vaddr"]
                gap_mem_size = first_sec["sh_addr"] - gap_mem_start
                if gap_mem_size > 0 and gap_file_start > 0:
                    flag = (
                        MemoryFlag.RW
                        if segment["p_flags"] & P_FLAGS.PF_W
                        else MemoryFlag.RX
                    )
                    block = MappedBlock(
                        gap_file_start,
                        gap_mem_start,
                        min(gap_file_size, gap_mem_size),
                        is_free=True,
                        flag=flag,
                    )
                    self.p.allocation_manager.add_block(block)
            if last_sec:
                gap_file_start = last_sec["sh_offset"] + last_sec["sh_size"]
                gap_file_size = segment_file_end - gap_file_start
                gap_mem_start = last_sec["sh_addr"] + last_sec["sh_size"]
                gap_mem_size = segment_mem_end - gap_mem_start
                if (
                    gap_mem_size > 0
                    and gap_file_size > 0
                    and segment_file_end > gap_file_start
                    and segment_mem_end > gap_mem_start
                ):
                    flag = (
                        MemoryFlag.RW
                        if segment["p_flags"] & P_FLAGS.PF_W
                        else MemoryFlag.RX
                    )
                    block = MappedBlock(
                        last_sec["sh_offset"] + last_sec["sh_size"],
                        gap_mem_start,
                        min(gap_file_size, gap_mem_size),
                        is_free=True,
                        flag=flag,
                    )
                    self.p.allocation_manager.add_block(block)

    def _find_space_between_segments(self) -> None:
        load_segments = sorted(
            (
                segment
                for segment in self._elf.iter_segments()
                if segment["p_type"] == "PT_LOAD"
            ),
            key=lambda x: x["p_offset"],
        )

        # Find Space Between Segments
        for curr, next in zip(load_segments, load_segments[1:]):
            if next["p_offset"] > (curr["p_offset"] + curr["p_filesz"]):
                block = FileBlock(
                    curr["p_offset"] + curr["p_filesz"],
                    next["p_offset"] - (curr["p_offset"] + curr["p_filesz"]),
                )
                self.p.allocation_manager.add_block(block)
            if next["p_vaddr"] > (curr["p_vaddr"] + curr["p_memsz"]):
                block = MemoryBlock(
                    curr["p_vaddr"] + curr["p_memsz"],
                    next["p_vaddr"] - (curr["p_vaddr"] + curr["p_memsz"]),
                )
                self.p.allocation_manager.add_block(block)

    def _add_end_of_file_block(self) -> None:
        load_segments = sorted(
            (
                segment
                for segment in self._elf.iter_segments()
                if segment["p_type"] == "PT_LOAD"
            ),
            key=lambda x: x["p_offset"],
        )

        if len(load_segments) == 0:
            return

        block = FileBlock(self.file_size, -1)
        self.p.allocation_manager.add_block(block)

        addr = load_segments[-1]["p_vaddr"] + load_segments[-1]["p_memsz"]
        # TODO: should we use the max_align of the segments?
        max_align = max([segment["p_align"] for segment in self._segments] + [0])
        addr = (addr + (max_align - 1)) & ~(max_align - 1)  # round up to max_align
        block = MemoryBlock(addr, -1)
        self.p.allocation_manager.add_block(block)

    def _init_memory_analysis(self) -> None:
        self._find_space_between_sections()
        self._find_space_between_segments()
        self._add_end_of_file_block()

        logger.debug(
            "_init_memory_analysis: "
            + "\n".join(map(str, self.p.allocation_manager.blocks.values()))
        )

        # # Is is possible to extend the existing segment?
        # for curr, next in zip(load_segments, load_segments[1:]):
        #     # mem
        #     mem_curr_end = curr['p_vaddr'] + curr['p_memsz']
        #     mem_curr_end_rounded = mem_curr_end + (curr['p_align'] - (mem_curr_end % curr['p_align'])) % curr['p_align']
        #     mem_next_start = next['p_vaddr']
        #     mem_max_extend_size = mem_curr_end_rounded - mem_curr_end
        #     if (mem_next_start - mem_curr_end_rounded) // curr['p_align'] > 0:
        #         mem_max_extend_size += (mem_next_start - mem_curr_end_rounded) // curr['p_align'] * curr['p_align']

        #     # file
        #     file_curr_end = curr['p_offset'] + curr['p_filesz']
        #     file_next_start = next['p_offset']
        #     max_extend_size = min(mem_max_extend_size, file_next_start - file_curr_end)

        #     if curr['p_flags'] & P_FLAGS.PF_R and curr['p_flags'] & P_FLAGS.PF_X:
        #         self.free_space['loaded'].append({
        #             "file_start": file_curr_end,
        #             "mem_start": mem_curr_end,
        #             "size": max_extend_size,
        #             "flag": MemoryFlag.RX,
        #             "callback": self._extend_segment
        #         })

    def finalize(self) -> None:
        self.p.allocation_manager.finalize()

        if len(self.p.allocation_manager.new_mapped_blocks) == 0:
            return

        load_segment_count = len(
            [segment for segment in self._segments if segment["p_type"] == "PT_LOAD"]
        )
        max_align = max([segment["p_align"] for segment in self._segments] + [0])

        # create new load segment for each new mapped block
        for block in self.p.allocation_manager.new_mapped_blocks:
            self._segments.append(
                Container(
                    **{
                        "p_type": "PT_LOAD",
                        "p_offset": block.file_addr,
                        "p_filesz": block.size,
                        "p_vaddr": block.mem_addr,
                        "p_paddr": block.mem_addr,
                        "p_memsz": block.size,
                        "p_flags": block.flag,
                        "p_align": max_align,  # TODO: what could be a good value for this?
                    }
                )
            )

        # sort segments by p_offset
        self._segments = sorted(self._segments, key=lambda x: x["p_offset"])

        # try to merge load segments if they are adjacent and have the same flags and same alignment
        # new size = sum of sizes of the two segments + gap between them
        while True:
            new_segments = []
            i = 0
            while i < len(self._segments) - 1:
                prev_seg = self._segments[i]
                next_seg = self._segments[i + 1]
                if (
                    prev_seg["p_type"] == next_seg["p_type"] == "PT_LOAD"
                    and prev_seg["p_offset"] + prev_seg["p_filesz"]
                    == next_seg["p_offset"]
                    and prev_seg["p_vaddr"] + prev_seg["p_memsz"] == next_seg["p_vaddr"]
                    and prev_seg["p_flags"] == next_seg["p_flags"]
                    and prev_seg["p_align"] == next_seg["p_align"]
                ):
                    new_segments.append(
                        Container(
                            **{
                                "p_type": "PT_LOAD",
                                "p_offset": prev_seg["p_offset"],
                                "p_filesz": prev_seg["p_filesz"]
                                + next_seg["p_filesz"]
                                + (
                                    next_seg["p_offset"]
                                    - (prev_seg["p_offset"] + prev_seg["p_filesz"])
                                ),
                                "p_vaddr": prev_seg["p_vaddr"],
                                "p_paddr": prev_seg["p_paddr"],
                                "p_memsz": prev_seg["p_memsz"]
                                + next_seg["p_memsz"]
                                + (
                                    next_seg["p_vaddr"]
                                    - (prev_seg["p_vaddr"] + prev_seg["p_memsz"])
                                ),
                                "p_flags": prev_seg["p_flags"],
                                "p_align": prev_seg["p_align"],
                            }
                        )
                    )
                    i += 2
                else:
                    new_segments.append(prev_seg)
                    i += 1
            if i == len(self._segments) - 1:
                new_segments.append(self._segments[i])
            if new_segments == self._segments:
                break
            self._segments = new_segments

        if (
            len(
                [
                    segment
                    for segment in self._segments
                    if segment["p_type"] == "PT_LOAD"
                ]
            )
            <= load_segment_count
        ):
            # just rebuild segment headers, it will be in place so we don't care if there is PHDR or not
            new_phdr = b""
            for segment in self._segments:
                new_phdr += self._elf.structs.Elf_Phdr.build(segment)
            self.p.binfmt_tool.update_binary_content(
                self._elf.header["e_phoff"], new_phdr
            )
            ehdr = self._elf.header
            ehdr["e_phnum"] = len(self._segments)
            new_ehdr = self._elf.structs.Elf_Ehdr.build(ehdr)
            self.p.binfmt_tool.update_binary_content(0, new_ehdr)
        else:
            # create new load segment for phdr (values are undetermined yet)
            phdr_load_segment = Container(
                **{
                    "p_type": "PT_LOAD",
                    "p_offset": 0,
                    "p_filesz": 0,
                    "p_vaddr": 0,
                    "p_paddr": 0,
                    "p_memsz": 0,
                    "p_flags": 0x4,
                    "p_align": 0x1000,
                }
            )
            self._segments.append(phdr_load_segment)

            # magic
            # TODO: should we use the max_align of the segments?
            load_segments_rounded = []
            first_load_segment = None
            for segment in self._segments:
                if segment["p_type"] == "PT_LOAD":
                    if first_load_segment is None:
                        first_load_segment = segment
                    load_segments_rounded.append(
                        (
                            # start of the segment, round down to multiple of max_align
                            (segment["p_vaddr"] - first_load_segment["p_vaddr"])
                            - (
                                (segment["p_vaddr"] - first_load_segment["p_vaddr"])
                                % max_align
                            ),
                            # end of the segment, round up to multiple of max_align
                            int(
                                (
                                    segment["p_vaddr"]
                                    + segment["p_memsz"]
                                    - first_load_segment["p_vaddr"]
                                    + max_align
                                    - 1
                                )
                                / max_align
                            )
                            * max_align,
                        )
                    )
            load_segments_rounded = sorted(load_segments_rounded, key=lambda x: x[0])

            # combine overlapping load segments
            while True:
                new_load_segments_rounded = []
                i = 0
                while i < len(load_segments_rounded) - 1:
                    prev_seg = load_segments_rounded[i]
                    next_seg = load_segments_rounded[i + 1]
                    if prev_seg[1] > next_seg[0]:  # two segments overlap
                        new_load_segments_rounded.append(
                            (prev_seg[0], next_seg[1])
                        )  # append combine of two segments
                        i += 2
                    else:
                        new_load_segments_rounded.append(
                            prev_seg
                        )  # append segment without overlap
                        i += 1
                if i == len(load_segments_rounded) - 1:
                    new_load_segments_rounded.append(
                        load_segments_rounded[i]
                    )  # append last segment if without overlapping
                if new_load_segments_rounded == load_segments_rounded:  # if no overlap
                    break
                load_segments_rounded = (
                    new_load_segments_rounded  # combined segments, run again
                )

            for prev_seg, next_seg in zip(
                load_segments_rounded[:-1], load_segments_rounded[1:]
            ):
                # TODO: should we use the max_align of the segments?
                potential_base = (
                    max(prev_seg[1], self.p.binfmt_tool.file_size) + (max_align - 1)
                ) & ~(max_align - 1)  # round up to max_align
                if next_seg[0] - potential_base > self._elf.header["e_phentsize"] * len(
                    self._segments
                ):  # if there is space between segments, put phdr here
                    phdr_start = potential_base
                    break
            else:
                phdr_start = load_segments_rounded[-1][
                    1
                ]  # otherwise put it after the last load segment
            # this is to workaround a weird issue in the dynamic linker of glibc
            # we want to make sure p_vaddr (phdr_start) == p_offset (len(ncontent))
            if phdr_start <= self.p.binfmt_tool.file_size:
                # TODO: should we use the max_align of the segments?
                # p_vaddr <= p_offset: pad the file (p_offset) to page size, and let p_vaddr = p_offset
                self.p.binfmt_tool.file_size = (
                    self.p.binfmt_tool.file_size + (max_align - 1)
                ) & ~(max_align - 1)  # round up to max_align
                phdr_start = self.p.binfmt_tool.file_size

            # update phdr segment and its corresponding load segment
            for segment in self._segments:
                if segment["p_type"] == "PT_PHDR" or segment == phdr_load_segment:
                    segment["p_filesz"] = self._elf.header["e_phentsize"] * len(
                        self._segments
                    )
                    segment["p_memsz"] = self._elf.header["e_phentsize"] * len(
                        self._segments
                    )
                    segment["p_offset"] = phdr_start
                    segment["p_vaddr"] = phdr_start + first_load_segment["p_vaddr"]
                    segment["p_paddr"] = phdr_start + first_load_segment["p_vaddr"]

            # create new phdr segment to be put in the new load segment above
            # make sure PHDR is the first segment
            self._segments = sorted(
                self._segments, key=lambda x: (x["p_type"] != "PT_PHDR", x["p_offset"])
            )
            new_phdr = b""
            for segment in self._segments:
                new_phdr += self._elf.structs.Elf_Phdr.build(segment)
            self.p.binfmt_tool.update_binary_content(phdr_start, new_phdr)

            ehdr = self._elf.header
            ehdr["e_phnum"] = len(self._segments)
            ehdr["e_phoff"] = phdr_start
            new_ehdr = self._elf.structs.Elf_Ehdr.build(ehdr)
            self.p.binfmt_tool.update_binary_content(0, new_ehdr)

    def save_binary(self, filename: str | None = None) -> None:
        self.updated_binary_content = self.updated_binary_content.ljust(
            self.file_size, b"\x00"
        )
        for update in self.file_updates:
            self.updated_binary_content = (
                self.updated_binary_content[: update["offset"]]
                + update["content"]
                + self.updated_binary_content[
                    update["offset"] + len(update["content"]) :
                ]
            )
        if filename is None:
            filename = f"{self.binary_path}.patched"
        with open(filename, "wb") as f:
            f.write(self.updated_binary_content)
        os.chmod(filename, 0o755)

    def update_binary_content(self, offset: int, new_content: bytes) -> None:
        logger.debug(
            f"Updating offset {hex(offset)} with content ({len(new_content)} bytes) {new_content.hex()}"
        )
        for update in self.file_updates:
            if offset >= update["offset"] and offset < update["offset"] + len(
                update["content"]
            ):
                raise ValueError(
                    f"Cannot update offset {hex(offset)} with content {new_content}, it overlaps with a previous update"
                )
        self.file_updates.append({"offset": offset, "content": new_content})
        if offset + len(new_content) > self.file_size:
            self.file_size = offset + len(new_content)

    def get_binary_content(self, offset: int, size: int) -> bytes:
        # check if it's in file_updates
        for update in self.file_updates:
            if offset >= update["offset"] and offset + size <= update["offset"] + len(
                update["content"]
            ):
                return update["content"][offset - update["offset"] :]
        # otherwise return from original binary
        return self.original_binary_content[offset : offset + size]

    def append_to_binary_content(self, new_content: bytes) -> None:
        self.file_updates.append({"offset": self.file_size, "content": new_content})
        self.file_size += len(new_content)



from __future__ import annotations

import logging

from .compiler import Compiler

logger = logging.getLogger(__name__)


class ClangArm(Compiler):
    def __init__(
        self, p, clang_version=15, compiler_flags: list[str] | None = None
    ) -> None:
        super().__init__(p)
        if compiler_flags is None:
            compiler_flags = []
        self._compiler = f"clang-{clang_version}"
        self._linker = f"ld.lld-{clang_version}"
        self._compiler_flags = compiler_flags

    def compile(
        self,
        code: str,
        base=0,
        symbols: dict[str, int] | None = None,
        extra_compiler_flags: list[str] | None = None,
        is_thumb=False,
        **kwargs,
    ) -> bytes:
        if symbols is None:
            symbols = {}
        if extra_compiler_flags is None:
            extra_compiler_flags = []
        if is_thumb:
            extra_compiler_flags += ["-mthumb"]
        else:
            extra_compiler_flags += ["-mno-thumb"]
        compiled = super().compile(
            code,
            base=base,
            symbols=symbols,
            extra_compiler_flags=extra_compiler_flags,
            **kwargs,
        )

        # FIXME: damn this is too hacky
        _symbols = {}
        _symbols.update(self.p.symbols)
        _symbols.update(self.p.binary_analyzer.get_all_symbols())
        _symbols.update(symbols)
        symbols = _symbols
        disasm = self.p.disassembler.disassemble(compiled, base=base, is_thumb=is_thumb)
        reassembled = b""
        for instr in disasm:
            if (
                is_thumb
                and instr["mnemonic"] == "bl"
                and int(instr["op_str"][1:], 0) in symbols.values()
            ):
                disasm_str = (
                    self.p.disassembler.to_asm_string(instr).replace("bl", "blx") + "\n"
                )
                reassembled += self.p.assembler.assemble(
                    disasm_str, base=instr["address"], is_thumb=is_thumb
                )
            elif (
                is_thumb
                and instr["mnemonic"] == "blx"
                and (int(instr["op_str"][1:], 0) + 1) in symbols.values()
            ):
                disasm_str = (
                    self.p.disassembler.to_asm_string(instr).replace("blx", "bl") + "\n"
                )
                reassembled += self.p.assembler.assemble(
                    disasm_str, base=instr["address"], is_thumb=is_thumb
                )
            elif (
                not is_thumb
                and instr["mnemonic"] == "bl"
                and (int(instr["op_str"][1:], 0) + 1) in symbols.values()
            ):
                disasm_str = (
                    self.p.disassembler.to_asm_string(instr).replace("bl", "blx") + "\n"
                )
                reassembled += self.p.assembler.assemble(
                    disasm_str, base=instr["address"], is_thumb=is_thumb
                )
            elif (
                not is_thumb
                and instr["mnemonic"] == "blx"
                and int(instr["op_str"][1:], 0) in symbols.values()
            ):
                disasm_str = (
                    self.p.disassembler.to_asm_string(instr).replace("blx", "bl") + "\n"
                )
                reassembled += self.p.assembler.assemble(
                    disasm_str, base=instr["address"], is_thumb=is_thumb
                )
            else:
                reassembled += compiled[
                    instr["address"] - base : instr["address"] - base + instr["size"]
                ]
        compiled = reassembled + compiled[len(reassembled) :]
        if len(compiled) % 2 != 0:
            compiled += b"\x00"
        return compiled



import logging
import os
import re
import subprocess
import tempfile

from ..assets.assets import Assets
from .assembler import Assembler

logger = logging.getLogger(__name__)


class PpcVle(Assembler):
    def __init__(self, p) -> None:
        super().__init__(p)
        self.assets_path = Assets("ppc_vle").path

    def _assemble(self, code: str, base=0, **kwargs) -> bytes:
        code = re.subn(r"\br(\d+)\b", r"\1", code)[0]

        if base is not None:
            # produce a list of {instr_offset: instr} pairs
            branch_instrs = {}
            instr_count = 0
            for line in code.splitlines():
                line = line.strip()
                if (
                    line.startswith(".")
                    or line.startswith("#")
                    or line == ""
                    or line.endswith(":")
                ):
                    continue
                # if line matches "b 0x*" or "bl 0x*", add it to the branch_instrs dict
                if re.match(r"b[a-z]* 0x[0-9a-fA-F]+", line):
                    branch_instrs[instr_count] = line
                instr_count += 1

            for i in range(instr_count):
                if i in branch_instrs:
                    branch_instrs[i] = (
                        branch_instrs[i].split(" ")[0]
                        + " "
                        + hex(int(branch_instrs[i].split(" ")[1], 16) - base - 4 * i)
                    )

            instr_count = 0
            for line_count, line in enumerate(code.splitlines()):
                if (
                    line.startswith(".")
                    or line.startswith("#")
                    or line == ""
                    or line.endswith(":")
                ):
                    continue
                if instr_count in branch_instrs:
                    code = code.splitlines()
                    code[line_count] = branch_instrs[instr_count]
                    code = "\n".join(code)
                instr_count += 1

        # set base address
        if base is not None:
            code = f".org {hex(base)}\n" + code

        with tempfile.TemporaryDirectory() as td:
            with open(os.path.join(td, "code.s"), "w") as f:
                f.write(code)
            try:
                subprocess.run(
                    [
                        os.path.join(self.assets_path, "powerpc-eabivle-as"),
                        "-o",
                        os.path.join(td, "obj.o"),
                        os.path.join(td, "code.s"),
                    ],
                    check=True,
                    capture_output=True,
                )
            except subprocess.CalledProcessError as e:
                logger.error(e.stderr.decode("utf-8"))
                raise e
            try:
                subprocess.run(
                    [
                        os.path.join(self.assets_path, "powerpc-eabivle-objcopy"),
                        "-O",
                        "binary",
                        "-j",
                        ".text",
                        os.path.join(td, "obj.o"),
                        os.path.join(td, "obj.bin"),
                    ],
                    check=True,
                    capture_output=True,
                )
            except subprocess.CalledProcessError as e:
                logger.error(e.stderr.decode("utf-8"))
                raise e
            with open(os.path.join(td, "obj.bin"), "rb") as f:
                if base != 0:
                    f.seek(base)
                binary = f.read()
                logger.debug(f"Assembled bytes: {bytes(binary).hex()}")
                return bytes(binary)



import logging
import os
import subprocess
import tempfile

from ..assets.assets import Assets
from .assembler import Assembler

logger = logging.getLogger(__name__)


class Bcc(Assembler):
    def __init__(self, p) -> None:
        super().__init__(p)
        self.assets_path = Assets("bcc").path

    def _assemble(self, code: str, base=0, **kwargs) -> bytes:
        with tempfile.TemporaryDirectory() as td:
            with open(os.path.join(td, "code.s"), "w") as f:
                f.write(f".org {hex(base)}\n")
                f.write(code)
            try:
                subprocess.run(
                    [
                        os.path.join(self.assets_path, "sparc-gaisler-elf-as"),
                        "-Aleon",
                        os.path.join(td, "code.s"),
                        "-o",
                        os.path.join(td, "obj.o"),
                    ],
                    check=True,
                    capture_output=True,
                )
            except subprocess.CalledProcessError as e:
                logger.error(e.stderr.decode("utf-8"))
                raise e
            try:
                subprocess.run(
                    [
                        os.path.join(self.assets_path, "sparc-gaisler-elf-objcopy"),
                        "-O",
                        "binary",
                        "-j",
                        ".text",
                        os.path.join(td, "obj.o"),
                        os.path.join(td, "obj.bin"),
                    ],
                    check=True,
                    capture_output=True,
                )
            except subprocess.CalledProcessError as e:
                logger.error(e.stderr.decode("utf-8"))
                raise e
            with open(os.path.join(td, "obj.bin"), "rb") as f:
                if base != 0:
                    f.seek(base)
                binary = f.read()
                logger.debug(f"Assembled bytes: {bytes(binary).hex()}")
                return bytes(binary)



class Amd64Info:
    nop_bytes = b"\x90"
    nop_size = 1
    jmp_asm = "jmp {dst}"
    jmp_size = 6
    alignment = 4
    bits = 64
    is_variable_length_isa = True
    instr_size = -1  # variable length
    call_asm = "call {dst}"
    pc_reg_names = ["rip"]
    save_context_asm = """
    push rax
    push rbx
    push rcx
    push rdx
    push rsi
    push rdi
    push rbp
    push rsp
    push r8
    push r9
    push r10
    push r11
    push r12
    push r13
    push r14
    push r15
    """
    restore_context_asm = """
    pop r15
    pop r14
    pop r13
    pop r12
    pop r11
    pop r10
    pop r9
    pop r8
    pop rsp
    pop rbp
    pop rdi
    pop rsi
    pop rdx
    pop rcx
    pop rbx
    pop rax
    """

    cc = {
        "Linux": ["rdi", "rsi", "rdx", "rcx", "r8", "r9"],
        "LinuxPreserveNone": [
            "r12",
            "r13",
            "r14",
            "r15",
            "rdi",
            "rsi",
            "rdx",
            "rcx",
            "r8",
            "r9",
            "r11",
            "rax",
        ],
        "Windows": ["rcx", "rdx", "r8", "r9"],
    }
    callee_saved = {"Linux": ["r12", "r13", "r14", "r15", "rbx", "rsp", "rbp"]}
    cc_float = {
        "Linux": ["xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5", "xmm6", "xmm7"]
    }
    callee_saved_float = {"Linux": []}

    float_types = {32: "float", 64: "double", 128: "__float128"}

    @property
    def regs(self):
        return list(self.subregisters.keys())

    @property
    def regs_float(self):
        return list(self.subregisters_float.keys())

    subregisters = {
        "rax": {
            64: ["rax"],
            32: ["eax"],
            16: ["ax"],
            # Note that the order of the children registers is important. Only the 0th
            # element of this list (al) is used when determining the calling convention.
            # That is, we can only use the following argument 'uint8_t al' in the
            # calling convention at the rax position. 'uint8_t ah' is NOT allowed.
            8: ["al", "ah"],
        },
        "rbx": {64: ["rbx"], 32: ["ebx"], 16: ["bx"], 8: ["bl", "bh"]},
        "rcx": {64: ["rcx"], 32: ["ecx"], 16: ["cx"], 8: ["cl", "ch"]},
        "rdx": {64: ["rdx"], 32: ["edx"], 16: ["dx"], 8: ["dl", "dh"]},
        "rsi": {64: ["rsi"], 32: ["esi"], 16: ["si"], 8: ["sil"]},
        "rdi": {64: ["rdi"], 32: ["edi"], 16: ["di"], 8: ["dil"]},
        "rbp": {64: ["rbp"], 32: ["ebp"], 16: ["bp"], 8: ["bpl"]},
        "rsp": {64: ["rsp"], 32: ["esp"], 16: ["sp"], 8: ["spl"]},
        "r8": {64: ["r8"], 32: ["r8d"], 16: ["r8w"], 8: ["r8b"]},
        "r9": {64: ["r9"], 32: ["r9d"], 16: ["r9w"], 8: ["r9b"]},
        "r10": {64: ["r10"], 32: ["r10d"], 16: ["r10w"], 8: ["r10b"]},
        "r11": {64: ["r11"], 32: ["r11d"], 16: ["r11w"], 8: ["r11b"]},
        "r12": {64: ["r12"], 32: ["r12d"], 16: ["r12w"], 8: ["r12b"]},
        "r13": {64: ["r13"], 32: ["r13d"], 16: ["r13w"], 8: ["r13b"]},
        "r14": {64: ["r14"], 32: ["r14d"], 16: ["r14w"], 8: ["r14b"]},
        "r15": {64: ["r15"], 32: ["r15d"], 16: ["r15w"], 8: ["r15b"]},
    }

    subregisters_float = {f"xmm{i}": {128: [f"xmm{i}"]} for i in range(0, 15 + 1)}



from patcherex2 import *

p = Patcherex("add")

new_add_func = """
int add(int a, int b) {
    return a * b;
}
"""

p.patches.append(ModifyFunctionPatch("add", new_add_func))

p.apply_patches()
p.save_binary("add_patched")



from __future__ import annotations


class Disassembler:
    def __init__(self, p) -> None:
        self.p = p

    def disassemble(self, input: bytes, base=0, **kwargs) -> None:
        raise NotImplementedError()

    def to_asm_string(self, insn: dict[str, int | str]) -> str:
        return "{} {}".format(insn["mnemonic"], insn["op_str"])



from __future__ import annotations

import capstone

from .disassembler import Disassembler


class Capstone(Disassembler):
    def __init__(self, arch: int, mode: int) -> None:
        self.cs = capstone.Cs(arch, mode)

    def disassemble(self, input: bytes, base=0, **kwargs) -> list[dict[str, int | str]]:
        cs_insns = self.cs.disasm(input, base)
        result = []
        for insn in cs_insns:
            result.append(
                {
                    "address": insn.address,
                    "size": insn.size,
                    "mnemonic": insn.mnemonic,
                    "op_str": insn.op_str,
                }
            )
        return result



import logging
import os
import tarfile
import tempfile
from pathlib import Path

import requests

logger = logging.getLogger(__name__)


class Assets:
    ASSETS_DIR = Path(__file__).parent
    ASSETS = {
        "bcc": {
            "url": "https://assets.patcherex.pursec.han.ac/bcc-2.2.4-gcc-linux64.tar.xz",
            "path": ASSETS_DIR / "bcc" / "bcc-2.2.4-gcc" / "bin",
        },
        "ppc_vle": {
            "url": "https://assets.patcherex.pursec.han.ac/powerpc-eabivle.tgz",
            "path": ASSETS_DIR / "ppc_vle" / "bin",
        },
        "llvm_recomp": {
            "url": "https://assets.patcherex.pursec.han.ac/llvm_recomp.tgz",
            "path": ASSETS_DIR / "llvm_recomp",
        },
    }

    def __init__(self, name: str) -> None:
        self.name = name
        self.url = self.ASSETS[name]["url"]
        self.path = self.ASSETS[name]["path"]
        if not os.path.exists(self.ASSETS_DIR / self.name):
            logger.info(f"{self.name} not found, downloading...")
            self.download()

    def download(self) -> None:
        r = requests.get(self.url)
        with tempfile.TemporaryDirectory() as td:
            with open(os.path.join(td, "asset.tgz"), "wb") as f:
                f.write(r.content)
            with tarfile.open(os.path.join(td, "asset.tgz")) as tar:
                # FIXME: better use filter here but it requires > py3.12. all tarball are manually verified to be safe so it's fine for now
                tar.extractall(path=self.ASSETS_DIR / self.name)



#!/usr/bin/env python

# ruff: noqa
import logging
import os
import shutil
import subprocess
import tempfile

import pytest

from patcherex2 import *

logging.getLogger("patcherex2").setLevel("DEBUG")


class Tests:
    @pytest.fixture(autouse=True, scope="class", params=["angr", "ghidra"])
    def setup(self, request):
        request.cls.bin_location = str(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "./test_binaries/armhf",
            )
        )
        request.cls.binary_analyzer = request.param

    def test_raw_file_patch_nopie(self):
        self.run_one(
            "printf_nopie",
            [ModifyRawBytesPatch(0x44C, b"No", addr_type="raw")],
            expected_output=b"No",
            expected_returnCode=0,
        )

    def test_raw_file_patch_pie(self):
        self.run_one(
            "printf_pie",
            [ModifyRawBytesPatch(0x5DF, b"No", addr_type="raw")],
            expected_output=b"No",
            expected_returnCode=0,
        )

    def test_raw_mem_patch_nopie(self):
        self.run_one(
            "printf_nopie",
            [ModifyRawBytesPatch(0x1044C, b"No")],
            expected_output=b"No",
            expected_returnCode=0,
        )

    def test_raw_mem_patch_pie(self):
        self.run_one(
            "printf_pie",
            [ModifyRawBytesPatch(0x5DF, b"No")],
            expected_output=b"No",
            expected_returnCode=0,
        )

    def test_modify_instruction_patch_nopie(self):
        self.run_one(
            "printf_nopie",
            [
                ModifyInstructionPatch(0x103E2, "ldr r1, [pc, #0x14]"),
                ModifyInstructionPatch(0x103E4, "add r1, pc"),
                ModifyInstructionPatch(0x103E6, "mov r0, r1"),
            ],
            expected_output=b"%s",
            expected_returnCode=0,
        )

    def test_modify_instruction_patch_pie(self):
        self.run_one(
            "printf_pie",
            [
                ModifyInstructionPatch(0x520, "mov r1, 0xb0"),
                ModifyInstructionPatch(0x524, "add r1, pc, r1"),
            ],
            expected_output=b"%s",
            expected_returnCode=0,
        )

    def test_insert_instruction_patch_nopie(self):
        instrs = """
            mov r7, 0x4
            mov r0, 0x1
            ldr r1, =0x1044C
            mov r2, 3
            svc 0
        """
        self.run_one(
            "printf_nopie",
            [InsertInstructionPatch(0x103EC, instrs)],
            expected_output=b"Hi\x00Hi",
            expected_returnCode=0,
        )

    def test_insert_instruction_patch_pie(self):
        instrs = """
            mov r7, 0x4
            mov r0, 0x1
            mov r1, lr
            ldr r2, =0xfffff000
            and r1, r1, r2
            ldr r2, =0x5df
            add r1, r1, r2
            mov r2, 3
            svc 0
        """
        self.run_one(
            "printf_pie",
            [InsertInstructionPatch(0x52C, instrs)],
            expected_output=b"Hi\x00Hi",
            expected_returnCode=0,
        )

    def test_insert_instruction_patch_2_nopie(self):
        instrs = """
            mov r7, 0x1
            mov r0, 0x32
            svc 0
        """
        self.run_one(
            "printf_nopie",
            [
                InsertInstructionPatch("return_0x32", instrs, is_thumb=True),
                ModifyInstructionPatch(0x103EC, "b {return_0x32}"),
            ],
            expected_returnCode=0x32,
        )

    @pytest.mark.skip(reason="difficult to do with pie")
    def test_insert_instruction_patch_2_pie(self):
        return_instrs = """
            mov r7, 0x1
            mov r0, 0x32
            svc 0
        """
        jump_instrs = """
            mov r0, pc
            ldr r1, =0xfff00000
            and r0, r0, r1
            ldr r1, ={return_0x32}
            add r0, r0, r1
            blx r0
        """
        self.run_one(
            "printf_pie",
            [
                InsertInstructionPatch("return_0x32", return_instrs, is_thumb=True),
                InsertInstructionPatch(0x52C, jump_instrs),
            ],
            expected_returnCode=0x32,
        )

    def test_remove_instruction_patch_nopie(self):
        self.run_one(
            "printf_nopie",
            [
                RemoveInstructionPatch(0x1044B, num_bytes=4),
            ],
            expected_output=b"\xf0\x20\xe3",
            expected_returnCode=0,
        )

    def test_remove_instruction_patch_pie(self):
        self.run_one(
            "printf_pie",
            [
                RemoveInstructionPatch(0x5DE, num_bytes=4),
            ],
            expected_output=b"\xf0\x20\xe3",
            expected_returnCode=0,
        )

    def test_modify_data_patch_nopie(self):
        self.run_one(
            "printf_nopie",
            [ModifyDataPatch(0x1044C, b"No")],
            expected_output=b"No",
            expected_returnCode=0,
        )

    def test_modify_data_patch_pie(self):
        self.run_one(
            "printf_pie",
            [ModifyDataPatch(0x5DF, b"No")],
            expected_output=b"No",
            expected_returnCode=0,
        )

    def test_insert_data_patch_nopie(self, tlen=5):
        p1 = InsertDataPatch("added_data", b"A" * tlen)
        instrs = """
            mov r7, 0x4
            mov r0, 0x1
            ldr r1, ={added_data}
            mov r2, %s
            svc 0
        """ % hex(tlen)
        p2 = InsertInstructionPatch(0x103EC, instrs)
        self.run_one(
            "printf_nopie",
            [p1, p2],
            expected_output=b"A" * tlen + b"Hi",
            expected_returnCode=0,
        )

    def test_insert_data_patch_pie(self, tlen=5):
        p1 = InsertDataPatch("added_data", b"A" * tlen)
        instrs = """
            mov r7, 0x4
            mov r0, 0x1
            mov r1, pc
            ldr r2, =0xfff00000
            and r1, r1, r2
            ldr r2, ={added_data}
            add r1, r1, r2
            mov r2, %s
            svc 0
        """ % hex(tlen)
        p2 = InsertInstructionPatch(0x52C, instrs)
        self.run_one(
            "printf_pie",
            [p1, p2],
            expected_output=b"A" * tlen + b"Hi",
            expected_returnCode=0,
        )

    def test_remove_data_patch_nopie(self):
        self.run_one(
            "printf_nopie",
            [RemoveDataPatch(0x1044D, 1)],
            expected_output=b"H",
            expected_returnCode=0,
        )

    def test_remove_data_patch_pie(self):
        self.run_one(
            "printf_pie",
            [RemoveDataPatch(0x5E0, 1)],
            expected_output=b"H",
            expected_returnCode=0,
        )

    def test_replace_function_patch(self):
        code = """
        int add(int a, int b){ for(;; b--, a+=2) if(b <= 0) return a; }
        """
        self.run_one(
            "replace_function_patch",
            [ModifyFunctionPatch(0x52C, code)],
            expected_output=b"70707070",
            expected_returnCode=0,
        )

    def test_replace_function_patch_with_function_reference(self):
        code = """
        extern int add(int, int);
        extern int subtract(int, int);
        int multiply(int a, int b){ for(int c = 0;; b = subtract(b, 1), c = subtract(c, a)) if(b <= 0) return c; }
        """
        self.run_one(
            "replace_function_patch",
            [ModifyFunctionPatch(0x588, code)],
            expected_output=b"-21-21",
            expected_returnCode=0,
        )

    def test_replace_function_patch_with_function_reference_and_rodata(self):
        code = """
        extern int printf(const char *format, ...);
        int multiply(int a, int b){ printf("%sWorld %s %s %s %d\\n", "Hello ", "Hello ", "Hello ", "Hello ", a * b);printf("%sWorld\\n", "Hello "); return a * b; }
        """
        self.run_one(
            "replace_function_patch",
            [ModifyFunctionPatch(0x588, code)],
            expected_output=b"Hello World Hello  Hello  Hello  21\nHello World\n2121",
            expected_returnCode=0,
        )

    def test_insert_function_patch(self):
        insert_code = """
        int min(int a, int b) { return (a < b) ? a : b; }
        """
        replace_code = """
        extern int min(int, int);
        int max(int a, int b) { return min(a, b); }
        """
        self.run_one(
            "replace_function_patch",
            [
                InsertFunctionPatch("min", insert_code),
                ModifyFunctionPatch(0x50C, replace_code),
            ],
            expected_output=b"2121212121",
            expected_returnCode=0,
        )

    def test_insert_instruction_patch_c(self):
        # Original computation was (n + 2) * 2 where n = 4
        # New computation inserts a factorial step at the beginning
        # The new computation is (n! + 2) * 2 where n=4
        instrs = """
        uint32_t n = r0;
        for (uint32_t i = 1; i < n; i++) {
            r0 *= i;
        }
        """

        config = InsertInstructionPatch.CConfig(
            scratch_regs=[
                "r1",
                "r2",
                "r3",
                "d0",
                "d1",
                "d2",
                "d3",
                "d4",
                "d5",
                "d6",
                "d7",
            ]
        )

        self.run_one(
            "iip_c",
            [InsertInstructionPatch(0x1062C, instrs, language="C", c_config=config)],
            expected_output=b"52",
            expected_returnCode=0,
        )

    def run_one(
        self,
        filename,
        patches,
        set_oep=None,
        inputvalue=None,
        expected_output=None,
        expected_returnCode=None,
        target_opts={},
    ):
        filepath = os.path.join(self.bin_location, filename)
        pipe = subprocess.PIPE

        with tempfile.TemporaryDirectory() as td:
            tmp_file = os.path.join(td, "patched")
            p = Patcherex(
                filepath,
                target_opts=target_opts | {"binary_analyzer": self.binary_analyzer},
            )
            for patch in patches:
                p.patches.append(patch)
            p.apply_patches()
            p.save_binary(tmp_file)
            p.shutdown()
            # os.system(f"readelf -hlS {tmp_file}")

            p = subprocess.Popen(
                ["qemu-arm", "-L", "/usr/arm-linux-gnueabihf", tmp_file],
                stdin=pipe,
                stdout=pipe,
                stderr=pipe,
            )
            res = p.communicate(inputvalue)
            if expected_output:
                if res[0] != expected_output:
                    pytest.fail(
                        f"AssertionError: {res[0]} != {expected_output}, binary dumped: {self.dump_file(tmp_file)}"
                    )
                # self.assertEqual(res[0], expected_output)
            if expected_returnCode:
                if p.returncode != expected_returnCode:
                    pytest.fail(
                        f"AssertionError: {p.returncode} != {expected_returnCode}, binary dumped: {self.dump_file(tmp_file)}"
                    )
                # self.assertEqual(p.returncode, expected_returnCode)

    def dump_file(self, file):
        shutil.copy(file, "/tmp/patcherex_failed_binary")
        return "/tmp/patcherex_failed_binary"



from elftools.construct.lib import Container

from ..components.allocation_managers.allocation_manager import (
    FileBlock,
    MemoryBlock,
)
from ..components.binfmt_tools.elf import ELF
from .elf_arm_linux import ElfArmLinux


class CustomElf(ELF):
    def _init_memory_analysis(self):
        """
        Information from NXP's MCUXpresso IDE:
        Flash is code, RAM4 is where data being loaded to

        * For additional code, we can just put them in the free flash space
        * For additional data, we also put them in the free flash space, but
        we need to update ResetISR to copy them to RAM4

        * The flasher (LinkServer) seems only care about segment headers, so
        we can safely ignore the section headers.

        * The IDE will strip the binary then call the flasher, and strip will
        remove the segment we added, so we need to add a fake corresponding
        section to make sure the segment is not removed.

        Type   | Name          | Alias | Location   | Size
        -------|---------------|-------|------------|----------
        Flash  | BOARD_FLASH   | Flash | 0x60000000 | 0x4000000
        RAM    | SRAM_DTC      | RAM   | 0x20000000 | 0x20000
        RAM    | SRAM_ITC      | RAM2  | 0x0        | 0x20000
        RAM    | SRAM_OC       | RAM3  | 0x20200000 | 0x20000
        RAM    | BOARD_SDRAM   | RAM4  | 0x80000000 | 0x1e00000
        RAM    | NCACHE_REGION | RAM5  | 0x81e00000 | 0x200000
        """

        # add free flash space to allocation manager
        flash_start = 0x60000000
        flash_end = 0x64000000
        highest_flash_addr = 0x60000000
        highest_file_offset = 0
        for segment in self._segments:
            seg_start = segment["p_paddr"]
            seg_end = segment["p_paddr"] + segment["p_memsz"]
            if (
                flash_start <= seg_start < flash_end
                and flash_start <= seg_end < flash_end
                and seg_end > highest_flash_addr
            ):
                highest_flash_addr = seg_end

            if segment["p_offset"] + segment["p_filesz"] > highest_file_offset:
                highest_file_offset = segment["p_offset"] + segment["p_filesz"]

        highest_file_offset = (highest_file_offset + 0xFFFF) & ~0xFFFF
        block = FileBlock(highest_file_offset, -1)
        self.p.allocation_manager.add_block(block)
        block = MemoryBlock(highest_flash_addr, -1)
        self.p.allocation_manager.add_block(block)

        return

    def finalize(self):
        self.p.allocation_manager.finalize()
        if len(self.p.allocation_manager.new_mapped_blocks) == 0:
            return

        max_align = max([segment["p_align"] for segment in self._segments] + [0])

        # create new load segment for each new mapped block
        for block in self.p.allocation_manager.new_mapped_blocks:
            self._segments.append(
                Container(
                    **{
                        "p_type": "PT_LOAD",
                        "p_offset": block.file_addr,
                        "p_filesz": block.size,
                        "p_vaddr": block.mem_addr,
                        "p_paddr": block.mem_addr,
                        "p_memsz": block.size,
                        "p_flags": block.flag,
                        "p_align": max_align,
                    }
                )
            )

            self._sections.append(
                Container(
                    **{
                        "sh_name": 0,
                        "sh_type": "SHT_PROGBITS",
                        "sh_flags": 2,
                        "sh_addr": block.mem_addr,
                        "sh_offset": block.file_addr,
                        "sh_size": block.size,
                        "sh_link": 0,
                        "sh_info": 0,
                        "sh_addralign": max_align,
                        "sh_entsize": 0,
                    }
                )
            )

        # sort segments by p_offset
        self._segments = sorted(self._segments, key=lambda x: x["p_offset"])

        # try to merge load segments if they are adjacent and have the same flags and same alignment
        # new size = sum of sizes of the two segments + gap between them
        while True:
            new_segments = []
            i = 0
            while i < len(self._segments) - 1:
                prev_seg = self._segments[i]
                next_seg = self._segments[i + 1]
                if (
                    prev_seg["p_type"] == next_seg["p_type"] == "PT_LOAD"
                    and prev_seg["p_offset"] + prev_seg["p_filesz"]
                    == next_seg["p_offset"]
                    and prev_seg["p_vaddr"] + prev_seg["p_memsz"] == next_seg["p_vaddr"]
                    and prev_seg["p_flags"] == next_seg["p_flags"]
                    and prev_seg["p_align"] == next_seg["p_align"]
                ):
                    new_segments.append(
                        Container(
                            **{
                                "p_type": "PT_LOAD",
                                "p_offset": prev_seg["p_offset"],
                                "p_filesz": prev_seg["p_filesz"]
                                + next_seg["p_filesz"]
                                + (
                                    next_seg["p_offset"]
                                    - (prev_seg["p_offset"] + prev_seg["p_filesz"])
                                ),
                                "p_vaddr": prev_seg["p_vaddr"],
                                "p_paddr": prev_seg["p_paddr"],
                                "p_memsz": prev_seg["p_memsz"]
                                + next_seg["p_memsz"]
                                + (
                                    next_seg["p_vaddr"]
                                    - (prev_seg["p_vaddr"] + prev_seg["p_memsz"])
                                ),
                                "p_flags": prev_seg["p_flags"],
                                "p_align": prev_seg["p_align"],
                            }
                        )
                    )
                    i += 2
                else:
                    new_segments.append(prev_seg)
                    i += 1
            if i == len(self._segments) - 1:
                new_segments.append(self._segments[i])
            if new_segments == self._segments:
                break
            self._segments = new_segments

        # generate new phdr at end of the file and update ehdr
        last_seg = sorted(self._segments, key=lambda x: x["p_offset"])[-1]
        phdr_start = last_seg["p_offset"] + last_seg["p_filesz"]
        new_phdr = b""
        for segment in self._segments:
            new_phdr += self._elf.structs.Elf_Phdr.build(segment)
        self.p.binfmt_tool.update_binary_content(phdr_start, new_phdr)

        ehdr = self._elf.header
        ehdr["e_phnum"] = len(self._segments)
        ehdr["e_phoff"] = phdr_start

        # generate new shdr at end of the file and update ehdr
        shdr_start = phdr_start + len(new_phdr)
        new_shdr = b""
        for section in self._sections:
            new_shdr += self._elf.structs.Elf_Shdr.build(section)
        self.p.binfmt_tool.update_binary_content(shdr_start, new_shdr)

        ehdr["e_shnum"] = len(self._sections)
        ehdr["e_shoff"] = shdr_start
        new_ehdr = self._elf.structs.Elf_Ehdr.build(ehdr)
        self.p.binfmt_tool.update_binary_content(0, new_ehdr)


class ElfArmMimxrt1052(ElfArmLinux):
    @staticmethod
    def detect_target(binary_path):
        return False

    def get_binfmt_tool(self, binfmt_tool):
        binfmt_tool = binfmt_tool or "default"
        if binfmt_tool == "default":
            return CustomElf(self.p, self.binary_path)
        raise NotImplementedError()



import logging

import keystone

from .assembler import Assembler

logger = logging.getLogger(__name__)


class Keystone(Assembler):
    def __init__(self, p, arch: int, mode: int) -> None:
        super().__init__(p)
        self.arch = arch
        self.mode = mode
        self.ks = keystone.Ks(arch, mode)

    def _assemble(self, code: str, base=0, **kwargs) -> bytes:
        try:
            binary, _ = self.ks.asm(code, base)
            logger.debug(f"Assembled bytes: {bytes(binary).hex()}")
            return bytes(binary)
        except Exception as e:
            raise Exception(
                f'Failed to assemble: """\n{code}\n"""\nat base: {hex(base)}'
            ) from e



from .data_patches import InsertDataPatch, ModifyDataPatch, RemoveDataPatch
from .dummy_patches import InsertLabelPatch, ModifyLabelPatch, RemoveLabelPatch
from .function_patches import (
    InsertFunctionPatch,
    ModifyFunctionPatch,
    RemoveFunctionPatch,
)
from .instruction_patches import (
    InsertInstructionPatch,
    ModifyInstructionPatch,
    RemoveInstructionPatch,
)
from .patch import Patch
from .raw_patches import ModifyRawBytesPatch

__all__ = [
    "ModifyDataPatch",
    "InsertDataPatch",
    "RemoveDataPatch",
    "InsertLabelPatch",
    "ModifyLabelPatch",
    "RemoveLabelPatch",
    "ModifyFunctionPatch",
    "InsertFunctionPatch",
    "RemoveFunctionPatch",
    "ModifyInstructionPatch",
    "InsertInstructionPatch",
    "RemoveInstructionPatch",
    "ModifyRawBytesPatch",
]


# Other Patches
class ModifyEntryPointPatch(Patch):
    def __init__(self, addr: int, parent=None) -> None:
        self.addr = addr
        super().__init__(parent)


# Complex Patches
class InsertFunctionWrapperPatch(Patch):
    def __init__(self, addr: int, wrapper_code: str, parent=None) -> None:
        self.addr = addr
        self.wrapper_code = wrapper_code
        super().__init__(parent)



from __future__ import annotations

import capstone

from .disassembler import Disassembler


class CapstoneArm(Disassembler):
    def __init__(self, p) -> None:
        super().__init__(p)
        self.cs_arm = capstone.Cs(
            capstone.CS_ARCH_ARM, capstone.CS_MODE_ARM + capstone.CS_MODE_LITTLE_ENDIAN
        )
        self.cs_thumb = capstone.Cs(
            capstone.CS_ARCH_ARM,
            capstone.CS_MODE_THUMB + capstone.CS_MODE_LITTLE_ENDIAN,
        )

    def disassemble(
        self, input: bytes, base=0, is_thumb=False, **kwargs
    ) -> list[dict[str, int | str]]:
        cs = self.cs_thumb if is_thumb else self.cs_arm
        cs_insns = cs.disasm(input, base)
        result = []
        for insn in cs_insns:
            result.append(
                {
                    "address": insn.address,
                    "size": insn.size,
                    "mnemonic": insn.mnemonic,
                    "op_str": insn.op_str,
                }
            )
        return result



import logging

logger = logging.getLogger(__name__)


class BinFmtTool:
    def __init__(self, p, binary_path: str) -> None:
        self.p = p
        self.binary_path = binary_path

    def _init_memory_analysis(self) -> None:
        raise NotImplementedError()

    def save_binary(self, filename=None) -> None:
        raise NotImplementedError()

    def update_binary_content(self, offset: str, new_content: bytes) -> None:
        raise NotImplementedError()

    def append_to_binary_content(self, new_content: bytes) -> None:
        raise NotImplementedError()



from __future__ import annotations

import io
import logging

from .binfmt_tool import BinFmtTool

logger = logging.getLogger(__name__)


class Binary(BinFmtTool):
    def __init__(self, p, binary_path: str) -> None:
        super().__init__(p, binary_path)
        self._file = open(binary_path, "rb")
        self.file_size = self._file.seek(0, io.SEEK_END)
        self._file.seek(0)
        self.file_updates = []

    def __del__(self) -> None:
        self._file.close()

    def _init_memory_analysis(self) -> None:
        pass

    def finalize(self) -> None:
        pass

    def save_binary(self, filename: str | None = None) -> None:
        if filename is None:
            filename = f"{self.binary_path}.patched"
        with open(filename, "wb") as f:
            self._file.seek(0)
            f.write(self._file.read())
            # apply the updates
            for update in self.file_updates:
                f.seek(update["offset"])
                f.write(update["content"])

    def update_binary_content(self, offset: int, new_content: bytes) -> None:
        logger.debug(
            f"Updating offset {hex(offset)} with content ({len(new_content)} bytes) {new_content.hex()}"
        )
        for update in self.file_updates:
            if offset >= update["offset"] and offset < update["offset"] + len(
                update["content"]
            ):
                raise ValueError(
                    f"Cannot update offset {hex(offset)} with content {new_content}, it overlaps with a previous update"
                )
        self.file_updates.append({"offset": offset, "content": new_content})
        if offset + len(new_content) > self.file_size:
            self.file_size = offset + len(new_content)

    def get_binary_content(self, offset: int, size: int) -> bytes:
        # FIXME: content partially in the file and partially in the updates (check other binfmt tools as well)
        # check if it's in the file updates
        for update in self.file_updates:
            if offset >= update["offset"] and offset + size <= update["offset"] + len(
                update["content"]
            ):
                return update["content"][
                    offset - update["offset"] : offset - update["offset"] + size
                ]
        # otherwise read from the file
        self._file.seek(offset)
        return self._file.read(size)

    def append_to_binary_content(self, new_content: bytes) -> None:
        self.file_updates.append({"offset": self.file_size, "content": new_content})
        self.file_size += len(new_content)



import logging
import re

import keystone

from .keystone import Keystone

logger = logging.getLogger(__name__)


class KeystoneSparc(Keystone):
    def _pre_assemble_hook(self, code: str, base=0) -> str:
        # Hack for sparc, remove this hack if keystone supports sparc relative addressing correctly
        # https://www.gaisler.com/doc/sparcv8.pdf
        # call (B.24.), all branches instr (B.21.) jumps to PC + 4ximm, so we need to convert absolute address to relative address
        if self.arch != keystone.KS_ARCH_SPARC:
            return code
        result = ""
        lineno = 0
        for line in code.splitlines():
            line = line.strip()
            if (
                line.startswith(".")
                or line.startswith("#")
                or line == ""
                or line.endswith(":")
            ):
                result += line + "\n"
                continue
            if re.match(r"(call|b|ba) 0x[0-9a-fA-F]+", line):
                instr, addr = line.split(" ")
                addr = int(addr, 16)
                disp = addr - base
                imm = disp >> 2
                logger.debug(
                    f"converting {line} to {instr} {hex(imm)} (base: {hex(base)})"
                )
                if instr == "call":
                    result += f"{instr} {hex(base + imm - lineno)}\n"
                elif instr == "b" or instr == "ba":
                    result += f"{instr} {hex(addr - (4 * lineno))}\n"
            else:
                result += line + "\n"
            lineno += 1
        return result



#!/usr/bin/env python

# ruff: noqa
import logging
import os
import shutil
import subprocess
import tempfile

import pytest

from patcherex2 import *

logging.getLogger("patcherex2").setLevel("DEBUG")


class Tests:
    @pytest.fixture(autouse=True, scope="class", params=["angr", "ghidra"])
    def setup(self, request):
        request.cls.bin_location = str(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "./test_binaries/aarch64",
            )
        )
        request.cls.binary_analyzer = request.param

    def test_raw_file_patch_nopie(self):
        self.run_one(
            "printf_nopie",
            [ModifyRawBytesPatch(0x640, b"No", addr_type="raw")],
            expected_output=b"No",
            expected_returnCode=0,
        )

    def test_raw_file_patch_pie(self):
        self.run_one(
            "printf_pie",
            [ModifyRawBytesPatch(0x7AB, b"No", addr_type="raw")],
            expected_output=b"No",
            expected_returnCode=0,
        )

    def test_raw_mem_patch_nopie(self):
        self.run_one(
            "printf_nopie",
            [ModifyRawBytesPatch(0x400640, b"No")],
            expected_output=b"No",
            expected_returnCode=0,
        )

    def test_raw_mem_patch_pie(self):
        self.run_one(
            "printf_pie",
            [ModifyRawBytesPatch(0x7AB, b"No")],
            expected_output=b"No",
            expected_returnCode=0,
        )

    def test_modify_instruction_patch_nopie(self):
        self.run_one(
            "printf_nopie",
            [ModifyInstructionPatch(0x400570, "add x1,x0,#0x648")],
            expected_output=b"%s",
            expected_returnCode=0,
        )

    def test_modify_instruction_patch_pie(self):
        self.run_one(
            "printf_pie",
            [ModifyInstructionPatch(0x778, "add x1,x1,#0x7a8")],
            expected_output=b"%s",
            expected_returnCode=0,
        )

    def test_insert_instruction_patch_nopie(self):
        instrs = """
            mov x8, 0x40
            mov x0, 0x1
            ldr x1, =0x400640
            mov x2, 0x3
            svc 0
        """
        self.run_one(
            "printf_nopie",
            [InsertInstructionPatch(0x400580, instrs)],
            expected_output=b"Hi\x00Hi",
            expected_returnCode=0,
        )

    def test_insert_instruction_patch_pie(self):
        instrs = """
            mov x8, 0x40
            mov x0, 0x1
            adrp x1, 0x0
            add x1, x1, #0x7ab
            mov x2, 0x3
            svc 0
        """
        self.run_one(
            "printf_pie",
            [InsertInstructionPatch(0x780, instrs)],
            expected_output=b"Hi\x00Hi",
            expected_returnCode=0,
        )

    def test_insert_instruction_patch_2_nopie(self):
        instrs = """
            mov x8, 0x5d
            mov x0, 0x32
            svc 0
        """
        self.run_one(
            "printf_nopie",
            [
                InsertInstructionPatch("return_0x32", instrs),
                ModifyInstructionPatch(0x400580, "b {return_0x32}"),
            ],
            expected_returnCode=0x32,
        )

    def test_insert_instruction_patch_2_pie(self):
        instrs = """
            mov x8, 0x5d
            mov x0, 0x32
            svc 0
        """
        self.run_one(
            "printf_pie",
            [
                InsertInstructionPatch("return_0x32", instrs),
                ModifyInstructionPatch(0x780, "b {return_0x32}"),
            ],
            expected_returnCode=0x32,
        )

    def test_remove_instruction_patch_nopie(self):
        self.run_one(
            "printf_nopie",
            [
                RemoveInstructionPatch(0x400641, num_bytes=4),
            ],
            expected_output=b"H\x1f\x20\x03\xd5",
            expected_returnCode=0,
        )

    def test_remove_instruction_patch_pie(self):
        self.run_one(
            "printf_pie",
            [
                RemoveInstructionPatch(0x7AC, num_bytes=4),
            ],
            expected_output=b"H\x1f\x20\x03\xd5",
            expected_returnCode=0,
        )

    def test_modify_data_patch_nopie(self):
        self.run_one(
            "printf_nopie",
            [ModifyDataPatch(0x400640, b"No")],
            expected_output=b"No",
            expected_returnCode=0,
        )

    def test_modify_data_patch_pie(self):
        self.run_one(
            "printf_pie",
            [ModifyDataPatch(0x7AB, b"No")],
            expected_output=b"No",
            expected_returnCode=0,
        )

    def test_insert_data_patch_nopie(self, tlen=5):
        p1 = InsertDataPatch("added_data", b"A" * tlen)
        instrs = """
            mov x8, 0x40
            mov x0, 0x1
            ldr x1, ={added_data}
            mov x2, %s
            svc 0
        """ % hex(tlen)
        p2 = InsertInstructionPatch(0x400580, instrs)
        self.run_one(
            "printf_nopie",
            [p1, p2],
            expected_output=b"A" * tlen + b"Hi",
            expected_returnCode=0,
        )

    def test_insert_data_patch_pie(self, tlen=5):
        p1 = InsertDataPatch("added_data", b"A" * tlen)
        instrs = """
            mov x8, 0x40
            mov x0, 0x1
            adrp x1, 0x0
            ldr x3, ={added_data}
            add x1, x1, x3
            mov x2, %s
            svc 0
        """ % hex(tlen)
        p2 = InsertInstructionPatch(0x780, instrs)
        self.run_one(
            "printf_pie",
            [p1, p2],
            expected_output=b"A" * tlen + b"Hi",
            expected_returnCode=0,
        )

    def test_remove_data_patch_nopie(self):
        self.run_one(
            "printf_nopie",
            [RemoveDataPatch(0x400641, 1)],
            expected_output=b"H",
            expected_returnCode=0,
        )

    def test_remove_data_patch_pie(self):
        self.run_one(
            "printf_pie",
            [RemoveDataPatch(0x7AC, 1)],
            expected_output=b"H",
            expected_returnCode=0,
        )

    def test_replace_function_patch(self):
        code = """
        int add(int a, int b){ for(;; b--, a+=2) if(b <= 0) return a; }
        """
        self.run_one(
            "replace_function_patch",
            [ModifyFunctionPatch(0x74C, code)],
            expected_output=b"70707070",
            expected_returnCode=0,
        )

    def test_replace_function_patch_with_function_reference(self):
        code = """
        extern int add(int, int);
        extern int subtract(int, int);
        int multiply(int a, int b){ for(int c = 0;; b = subtract(b, 1), c = subtract(c, a)) if(b <= 0) return c; }
        """
        self.run_one(
            "replace_function_patch",
            [ModifyFunctionPatch(0x7D4, code)],
            expected_output=b"-21-21",
            expected_returnCode=0,
        )

    def test_replace_function_patch_with_function_reference_and_rodata(self):
        code = """
        extern int printf(const char *format, ...);
        int multiply(int a, int b){ printf("%sWorld %s %s %s %d\\n", "Hello ", "Hello ", "Hello ", "Hello ", a * b);printf("%sWorld\\n", "Hello "); return a * b; }
        """
        self.run_one(
            "replace_function_patch",
            [ModifyFunctionPatch(0x7D4, code)],
            expected_output=b"Hello World Hello  Hello  Hello  21\nHello World\n2121",
            expected_returnCode=0,
        )

    def test_insert_function_patch(self):
        insert_code = """
        int min(int a, int b) { return (a < b) ? a : b; }
        """
        replace_code = """
        extern int min(int, int);
        int max(int a, int b) { return min(a, b); }
        """
        self.run_one(
            "replace_function_patch",
            [
                InsertFunctionPatch("min", insert_code),
                ModifyFunctionPatch(0x724, replace_code),
            ],
            expected_output=b"2121212121",
            expected_returnCode=0,
        )

    def test_insert_instruction_patch_c(self):
        # Original computation was (n + 2) * 2 where n = 4
        # New computation inserts a factorial step at the beginning
        # The new computation is (n! + 2) * 2 where n=4
        instrs = """
        uint32_t n = x0;
        for (uint32_t i = 1; i < n; i++) {
            x0 *= i;
        }
        """

        config = InsertInstructionPatch.CConfig(
            scratch_regs=[
                "x1",
                "x2",
                "x3",
                "x4",
                "x5",
                "x6",
                "x7",
                "x8",
                "x9",
                "x10",
                "x11",
                "x12",
                "x13",
                "x14",
                "x15",
                "v0",
                "v1",
                "v2",
                "v3",
                "v4",
                "v5",
                "v6",
                "v7",
            ]
        )

        self.run_one(
            "iip_c",
            [InsertInstructionPatch(0x760, instrs, language="C", c_config=config)],
            expected_output=b"52",
            expected_returnCode=0,
        )

    def test_insert_instruction_patch_c_subreg(self):
        # TODO: Currently this test fails due to the following missing cle relocation:
        # WARNING  | 2024-05-08 13:58:10,162 | cle.backends.elf.relocation | Unknown reloc 299 on AARCH64

        # Original computation was (n + 2) * 2 where n = 4
        # New computation inserts a factorial step at the beginning
        # The new computation is (n! + 2) * 2 where n=4
        instrs = """
        uint32_t n = w0;
        for (uint32_t i = 1; i < n; i++) {
            w0 *= i;
        }
        """

        config = InsertInstructionPatch.CConfig(
            scratch_regs=[
                "x1",
                "x2",
                "x3",
                "x4",
                "x5",
                "x6",
                "x7",
                "x8",
                "x9",
                "x10",
                "x11",
                "x12",
                "x13",
                "x14",
                "x15",
                "v0",
                "v1",
                "v2",
                "v3",
                "v4",
                "v5",
                "v6",
                "v7",
            ],
            regs_sort=["w0"],
        )

        self.run_one(
            "iip_c",
            [InsertInstructionPatch(0x760, instrs, language="C", c_config=config)],
            expected_output=b"52",
            expected_returnCode=0,
        )

    def test_insert_instruction_patch_c_asm_header(self):
        asm_header = "mov x12, x29"

        # Original computation computed the area as pi * radius
        # Here our micropatch loops over the areas array and multiplies by another radius to fix the bug
        instrs = """
        int num_radii = *((int *) (x12 - 0x2c));
        float *areas = *((float **) (x12 - 0x10));
        float *radii = *((float **) (x12 - 0x28));
        for (int i = 0; i < num_radii; i++) {
            areas[i] *= radii[i];
        }
        """

        config = InsertInstructionPatch.CConfig(
            asm_header=asm_header,
            scratch_regs=[
                "x1",
                "x3",
                "x4",
                "x5",
                "x6",
                "x7",
                "x8",
                "x9",
                "x10",
                "x11",
                "x13",
                "x14",
                "x15",
                "v0",
                "v1",
                "v2",
                "v3",
                "v4",
                "v5",
                "v6",
                "v7",
            ],
        )

        expected_output = b"".join(
            [
                b"The area of the circle with radius 1.500000 is 7.065000\n",
                b"The area of the circle with radius 2.000000 is 12.560000\n",
                b"The area of the circle with radius 4.300000 is 58.058605\n",
            ]
        )

        self.run_one(
            "iip_c_asm_header",
            [InsertInstructionPatch(0xA20, instrs, language="C", c_config=config)],
            expected_output=expected_output,
            expected_returnCode=0,
            target_opts={"compiler": "clang19"},
        )

    def test_insert_instruction_patch_c_asm_header2(self):
        asm_header = "mov x12, x29"

        # Original computation computed the area as pi * radius
        # Here our micropatch loops over the areas array and multiplies by another radius to fix the bug
        instrs = """
        int num_radii = 3;
        float *areas = *((float **) (x12 - 0x10));
        float *radii = *((float **) (x12 - 0x28));
        for (int i = 0; i < num_radii; i++) {
            areas[i] *= radii[i];
        }
        """

        config = InsertInstructionPatch.CConfig(
            asm_header=asm_header,
            scratch_regs=[
                "x1",
                "x3",
                "x4",
                "x5",
                "x6",
                "x7",
                "x8",
                "x9",
                "x10",
                "x11",
                "x13",
                "x14",
                "x15",
                "v0",
                "v1",
                "v2",
                "v3",
                "v4",
                "v5",
                "v6",
                "v7",
            ],
        )

        expected_output = b"".join(
            [
                b"The area of the circle with radius 1.500000 is 7.065000\n",
                b"The area of the circle with radius 2.000000 is 12.560000\n",
                b"The area of the circle with radius 4.300000 is 58.058605\n",
            ]
        )

        self.run_one(
            "iip_c_asm_header",
            [InsertInstructionPatch(0xA20, instrs, language="C", c_config=config)],
            expected_output=expected_output,
            expected_returnCode=0,
            target_opts={"compiler": "clang19"},
        )

    def test_insert_instruction_patch_c_float(self):
        # Original computation calculated the square magnitude of a 3D vector as x^2 + y^2
        # Here we insert an additional step to fix the calculation to be x^2 + y^2 + z^2
        instrs = """
        s0 += s2 * s2;
        """

        config = InsertInstructionPatch.CConfig(
            scratch_regs=[
                "x1",
                "x2",
                "x3",
                "x4",
                "x5",
                "x6",
                "x7",
                "x8",
                "x9",
                "x10",
                "x11",
                "x12",
                "x13",
                "x14",
                "x15",
                "v3",
                "v4",
                "v5",
                "v6",
                "v7",
            ],
            regs_sort=["s0", "s1", "s2"],
        )

        expected_output = b"".join(
            [
                b"The square magnitude of the vector (0.000000, 0.000000, 0.000000) is 0.000000\n",
                b"The square magnitude of the vector (1.000000, 2.000000, 3.000000) is 14.000000\n",
                b"The square magnitude of the vector (-20.000000, 33.200001, 5.200000) is 1529.280029\n",
                b"The square magnitude of the vector (3.000000, 4.000000, 0.000000) is 25.000000\n",
            ]
        )

        self.run_one(
            "iip_c_float",
            [InsertInstructionPatch(0x774, instrs, language="C", c_config=config)],
            expected_output=expected_output,
            expected_returnCode=0,
            target_opts={"compiler": "clang19"},
        )

    def run_one(
        self,
        filename,
        patches,
        set_oep=None,
        inputvalue=None,
        expected_output=None,
        expected_returnCode=None,
        target_opts={},
    ):
        filepath = os.path.join(self.bin_location, filename)
        pipe = subprocess.PIPE

        with tempfile.TemporaryDirectory() as td:
            tmp_file = os.path.join(td, "patched")
            p = Patcherex(
                filepath,
                target_opts=target_opts | {"binary_analyzer": self.binary_analyzer},
            )
            for patch in patches:
                p.patches.append(patch)
            p.apply_patches()
            p.save_binary(tmp_file)
            p.shutdown()
            # os.system(f"readelf -hlS {tmp_file}")

            p = subprocess.Popen(
                ["qemu-aarch64", "-L", "/usr/aarch64-linux-gnu", tmp_file],
                stdin=pipe,
                stdout=pipe,
                stderr=pipe,
            )
            res = p.communicate(inputvalue)
            if expected_output:
                if res[0] != expected_output:
                    pytest.fail(
                        f"AssertionError: {res[0]} != {expected_output}, binary dumped: {self.dump_file(tmp_file)}"
                    )
                # self.assertEqual(res[0], expected_output)
            if expected_returnCode:
                if p.returncode != expected_returnCode:
                    pytest.fail(
                        f"AssertionError: {p.returncode} != {expected_returnCode}, binary dumped: {self.dump_file(tmp_file)}"
                    )
                # self.assertEqual(p.returncode, expected_returnCode)

    def dump_file(self, file):
        shutil.copy(file, "/tmp/patcherex_failed_binary")
        return "/tmp/patcherex_failed_binary"



from __future__ import annotations

from patcherex2.patches import Patch


class PatchManager:
    def __init__(self) -> None:
        self.patches = []
        self.analyzed = False

    def add_patch(self, patch: Patch) -> None:
        self.analyzed = False
        self.patches.append(patch)

    def add_patches(self, patches: list[Patch]) -> None:
        for patch in patches:
            self.add_patch(patch)

    def export_patches(self, filename: str) -> None:
        raise NotImplementedError()

    def import_patches(self, filename: str) -> None:
        raise NotImplementedError()

    def analyze_patches(self, ignore_conflicts=False) -> None:
        raise NotImplementedError()

    def apply_patches(self, best_effort=False) -> None:
        raise NotImplementedError()



from __future__ import annotations

import logging
import traceback

import angr
from archinfo import ArchARM

from .binary_analyzer import BinaryAnalyzer

logger = logging.getLogger(__name__)


class Angr(BinaryAnalyzer):
    def __init__(self, binary_path: str, **kwargs) -> None:
        self.binary_path = binary_path
        # self.use_pickle = kwargs.pop("use_pickle", False) # TODO: implement this
        self.angr_kwargs = kwargs.pop("angr_kwargs", {})
        self.angr_cfg_kwargs = kwargs.pop("angr_cfg_kwargs", {})
        self._p = None
        self._cfg = None
        self._load_base = None

    @property
    def load_base(self) -> int:
        if self._load_base is None:
            self._load_base = self.p.loader.main_object.mapped_base
        return self._load_base

    def normalize_addr(self, addr: int) -> int:
        if self.p.loader.main_object.pic:
            return addr - self.load_base
        return addr

    def denormalize_addr(self, addr: int) -> int:
        if self.p.loader.main_object.pic:
            return addr + self.load_base
        return addr

    @property
    def p(self) -> angr.Project:
        if self._p is None:
            logger.info("Loading binary with angr")
            if "load_options" not in self.angr_kwargs:
                self.angr_kwargs["load_options"] = {"auto_load_libs": False}
            self._p = angr.Project(self.binary_path, **self.angr_kwargs)
            logger.info("Loaded binary with angr")
        return self._p

    @property
    def cfg(self) -> angr.analyses.cfg.cfg_fast.CFGFast:
        if self._cfg is None:
            logger.info("Generating CFG with angr")
            if "normalize" not in self.angr_cfg_kwargs:
                # NOTE: This will split basic blocks if another block jumps to the middle of the block
                self.angr_cfg_kwargs["normalize"] = True
            self._cfg = self.p.analyses.CFGFast(**self.angr_cfg_kwargs)
            logger.info("Generated CFG with angr")
        return self._cfg

    def mem_addr_to_file_offset(self, addr: int) -> int:
        addr = self.denormalize_addr(addr)
        file_addr = self.p.loader.main_object.addr_to_offset(addr)
        if file_addr is None:
            logger.error(
                f"Cannot convert memory address {hex(addr)} to file offset, will use the memory address instead"
            )
            return addr
        return file_addr

    def get_basic_block(self, addr: int) -> dict[str, int | list[int]]:
        # NOTE: angr splits basic blocks at call instructions, so we need to handle this
        if self.is_thumb(addr) and addr % 2 == 0:
            addr += 1
        addr = self.denormalize_addr(addr)

        try:
            func = self.p.kb.functions.function(
                self.cfg.model.get_any_node(addr, anyaddr=True).function_address
            )
            ri = self.p.analyses.RegionIdentifier(func)
            graph = ri._graph.copy()
            ri._make_supergraph(graph)

            for multinode in graph.nodes():
                nodes = multinode.nodes if hasattr(multinode, "nodes") else [multinode]
                start = multinode.addr
                size = sum(node.size for node in nodes)
                end = start + size

                instr_addrs = [
                    instr_addr
                    for node in nodes
                    for instr_addr in func.get_block(node.addr).instruction_addrs
                ]

                if addr in instr_addrs:
                    return {
                        "start": self.normalize_addr(start),
                        "end": self.normalize_addr(end),
                        "size": size,
                        "instruction_addrs": [
                            self.normalize_addr(instr_addr)
                            - (
                                1
                                if self.is_thumb(self.normalize_addr(instr_addr))
                                else 0
                            )
                            for instr_addr in instr_addrs
                        ],
                    }
        except Exception:
            logger.error(
                f"angr RegionIdentifier failed for function containing {hex(addr)}, falling back to use cfg nodes\n{traceback.format_exc()}"
            )
            bb = None
            for node in self.cfg.model.nodes():
                if addr in node.instruction_addrs:
                    bb = node
                    break
            assert bb is not None
            return {
                "start": self.normalize_addr(bb.addr),
                "end": self.normalize_addr(bb.addr + bb.size),
                "size": bb.size,
                "instruction_addrs": [
                    self.normalize_addr(addr)
                    - (1 if self.is_thumb(self.normalize_addr(addr)) else 0)
                    for addr in bb.instruction_addrs
                ],
            }

        raise Exception(f"Cannot find a block containing address {hex(addr)}")

    def get_instr_bytes_at(self, addr: int, num_instr=1) -> angr.Block:
        addr += 1 if self.is_thumb(addr) else 0
        addr = self.denormalize_addr(addr)
        # TODO: Special handling for delay slot, when there is a call instr with delay slot
        # angr will return both instrs, even when num_instr is 1
        return self.p.factory.block(addr, num_inst=num_instr).bytes

    def get_unused_funcs(self) -> list[dict[str, int]]:
        logger.info("Getting unused functions with angr")
        unused_funcs = []
        assert self.cfg is not None
        for func in self.p.kb.functions.values():
            if func.size == 0:
                continue
            for dst, _ in self.p.kb.xrefs.xrefs_by_dst.items():
                if dst == func.addr:
                    break
            else:
                unused_funcs.append(
                    {
                        "addr": self.normalize_addr(func.addr)
                        - (1 if self.is_thumb(self.normalize_addr(func.addr)) else 0),
                        "size": func.size,
                    }
                )
        return unused_funcs

    def get_all_symbols(self) -> dict[str, int]:
        assert self.cfg is not None
        logger.info("Getting all symbols with angr")
        symbols = {}
        for symbol in self.p.loader.main_object.symbols:
            if not symbol.name or not symbol.is_function:
                continue
            symbols[symbol.name] = self.normalize_addr(symbol.rebased_addr)
        for func in self.p.kb.functions.values():
            # make it compatible with old angr versions
            if func.is_simprocedure or (getattr(func, "is_alignment", func.alignment)):
                continue
            symbols[func.name] = self.normalize_addr(func.addr)
        return symbols

    def get_function(self, name_or_addr: int | str) -> dict[str, int] | None:
        assert self.cfg is not None
        if isinstance(name_or_addr, (str, int)):
            if isinstance(name_or_addr, int):
                name_or_addr += 1 if self.is_thumb(name_or_addr) else 0
                name_or_addr = self.denormalize_addr(name_or_addr)
            if name_or_addr in self.p.kb.functions:
                func = self.p.kb.functions[name_or_addr]
                return {
                    "addr": self.normalize_addr(func.addr)
                    - (1 if self.is_thumb(self.normalize_addr(func.addr)) else 0),
                    "size": func.size,
                }
            return None
        else:
            raise Exception(f"Invalid type for name_or_addr: {type(name_or_addr)}")

    def is_thumb(self, addr: int) -> bool:
        if not isinstance(self.p.arch, ArchARM):
            return False
        addr = self.denormalize_addr(addr)

        for node in self.cfg.model.nodes():
            if addr in node.instruction_addrs:
                return node.thumb
        else:
            if addr % 2 == 0:
                return self.is_thumb(self.normalize_addr(addr + 1))
            else:
                logger.error(f"Cannot find a block containing address {hex(addr)}")
                return False



import copy

from elftools.construct.lib import Container

from ..components.allocation_managers.allocation_manager import (
    AllocationManager,
    FileBlock,
    MappedBlock,
    MemoryBlock,
    MemoryFlag,
)
from ..components.binfmt_tools.elf import ELF
from ..patches import InsertInstructionPatch
from .elf_arm_linux import ElfArmLinux


class FlashBlock(MemoryBlock):
    pass


class RamBlock(MemoryBlock):
    pass


class CustomAllocationManager(AllocationManager):
    def _create_new_mapped_block(
        self, size: int, flag=MemoryFlag.RWX, align=0x1
    ) -> bool:
        file_addr = None
        virtual_mem_addr = None
        load_mem_addr = None
        for block in self.blocks[FileBlock]:
            if block.size == -1:
                file_addr = block.addr
                block.addr += 0x10000
        if flag == MemoryFlag.RW:
            for block in self.blocks[RamBlock]:
                if block.size == -1:
                    if self.p.binfmt_tool.__class__.__name__ == "ELF":
                        max_seg_align = max(
                            [
                                segment["p_align"]
                                for segment in self.p.binfmt_tool._segments
                            ]
                            + [0]
                        )
                        virtual_mem_addr = (
                            block.addr + (file_addr - block.addr) % max_seg_align
                        )
                    else:
                        virtual_mem_addr = (
                            block.addr + (file_addr - block.addr) % 0x1000
                        )
                    block.addr = virtual_mem_addr + 0x10000
            for block in self.blocks[FlashBlock]:
                if block.size == -1:
                    if self.p.binfmt_tool.__class__.__name__ == "ELF":
                        max_seg_align = max(
                            [
                                segment["p_align"]
                                for segment in self.p.binfmt_tool._segments
                            ]
                            + [0]
                        )
                        load_mem_addr = (
                            block.addr + (file_addr - block.addr) % max_seg_align
                        )
                    else:
                        load_mem_addr = block.addr + (file_addr - block.addr) % 0x1000
                    block.addr = load_mem_addr + 0x10000
            if file_addr and load_mem_addr and virtual_mem_addr:
                block = MappedBlock(
                    file_addr,
                    virtual_mem_addr,
                    0x10000,
                    is_free=True,
                    flag=flag,
                    load_mem_addr=load_mem_addr,
                )
                self.add_block(copy.deepcopy(block))
                self.new_mapped_blocks.append(copy.deepcopy(block))
                return True
        elif flag == MemoryFlag.RX:
            for block in self.blocks[FlashBlock]:
                if block.size == -1:
                    # NOTE: mem_addr % p_align should equal to file_addr % p_align
                    # Check `man elf` and search for `p_align` for more information
                    # FIXME: shouldn't do any assumption on component type, reimpl in a better way
                    # FIXME: even worse, importing ELF will cause circular import
                    # TODO: consider merge allocation_manager and binfmt_tool into one component
                    if self.p.binfmt_tool.__class__.__name__ == "ELF":
                        max_seg_align = max(
                            [
                                segment["p_align"]
                                for segment in self.p.binfmt_tool._segments
                            ]
                            + [0]
                        )
                        load_mem_addr = (
                            block.addr + (file_addr - block.addr) % max_seg_align
                        )
                    else:
                        load_mem_addr = block.addr + (file_addr - block.addr) % 0x1000
                    block.addr = load_mem_addr + 0x10000
            if file_addr and load_mem_addr:
                block = MappedBlock(
                    file_addr,
                    load_mem_addr,
                    0x10000,
                    is_free=True,
                    flag=flag,
                )
                self.add_block(copy.deepcopy(block))
                self.new_mapped_blocks.append(copy.deepcopy(block))
                return True
        else:
            raise NotImplementedError("Unknown MemoryFlag")
        return False


class CustomElf(ELF):
    def __init__(self, p, binary_path, **kwargs):
        assert (
            "flash_start" in kwargs
            and "flash_end" in kwargs
            and "ram_start" in kwargs
            and "ram_end" in kwargs
            and "insert_points" in kwargs
        )
        self.flash_start = kwargs["flash_start"]
        self.flash_end = kwargs["flash_end"]
        self.ram_start = kwargs["ram_start"]
        self.ram_end = kwargs["ram_end"]
        self.insert_points = kwargs["insert_points"]
        assert isinstance(self.insert_points, list)

        super().__init__(p, binary_path)

    def _init_memory_analysis(self):
        highest_flash_addr = self.flash_start
        highest_ram_addr = self.ram_start
        highest_file_offset = 0
        for segment in self._segments:
            seg_start = segment["p_vaddr"]
            seg_end = segment["p_vaddr"] + segment["p_memsz"]
            if (
                self.flash_start <= seg_start < self.flash_end
                and self.flash_start <= seg_end < self.flash_end
                and seg_end > highest_flash_addr
            ):
                highest_flash_addr = seg_end
            if (
                self.ram_start <= seg_start < self.ram_end
                and self.ram_start <= seg_end < self.ram_end
                and seg_end > highest_ram_addr
            ):
                highest_ram_addr = seg_end

            if segment["p_offset"] + segment["p_filesz"] > highest_file_offset:
                highest_file_offset = segment["p_offset"] + segment["p_filesz"]

        highest_file_offset = (highest_file_offset + 0xFFFF) & ~0xFFFF
        block = FileBlock(highest_file_offset, -1)
        self.p.allocation_manager.add_block(block)
        block = RamBlock(highest_ram_addr, -1)
        self.p.allocation_manager.add_block(block)
        block = FlashBlock(highest_flash_addr, -1)
        self.p.allocation_manager.add_block(block)

    def finalize(self):
        self.p.allocation_manager.finalize()
        if len(self.p.allocation_manager.new_mapped_blocks) == 0:
            return

        max_align = max([segment["p_align"] for segment in self._segments] + [0])

        copy_to_ram = ""
        # create new load segment for each new mapped block
        for block in self.p.allocation_manager.new_mapped_blocks:
            if block.mem_addr != block.load_mem_addr:
                copy_to_ram += f"""
ldr r0, ={hex(block.load_mem_addr)}
ldr r1, ={hex(block.mem_addr)}
ldr r2, ={hex(block.size)}
copy:
ldrb r3, [r0], #1
strb r3, [r1], #1
subs r2, r2, #1
bne copy
"""
        for insert_point in self.insert_points:
            InsertInstructionPatch(insert_point, copy_to_ram, save_context=True).apply(
                self.p
            )
        self.p.allocation_manager.finalize()
        # create new load segment for each new mapped block
        for block in self.p.allocation_manager.new_mapped_blocks:
            self._segments.append(
                Container(
                    **{
                        "p_type": "PT_LOAD",
                        "p_offset": block.file_addr,
                        "p_filesz": block.size,
                        "p_vaddr": block.mem_addr,
                        "p_paddr": block.load_mem_addr,
                        "p_memsz": block.size,
                        "p_flags": block.flag,
                        "p_align": max_align,
                    }
                )
            )

            self._sections.append(
                Container(
                    **{
                        "sh_name": 0,
                        "sh_type": "SHT_PROGBITS",
                        "sh_flags": 2,
                        "sh_addr": block.mem_addr,
                        "sh_offset": block.file_addr,
                        "sh_size": block.size,
                        "sh_link": 0,
                        "sh_info": 0,
                        "sh_addralign": max_align,
                        "sh_entsize": 0,
                    }
                )
            )

        # sort segments by p_offset
        self._segments = sorted(self._segments, key=lambda x: x["p_offset"])

        # try to merge load segments if they are adjacent and have the same flags and same alignment
        # new size = sum of sizes of the two segments + gap between them
        while True:
            new_segments = []
            i = 0
            while i < len(self._segments) - 1:
                prev_seg = self._segments[i]
                next_seg = self._segments[i + 1]
                if (
                    prev_seg["p_type"] == next_seg["p_type"] == "PT_LOAD"
                    and prev_seg["p_offset"] + prev_seg["p_filesz"]
                    == next_seg["p_offset"]
                    and prev_seg["p_vaddr"] + prev_seg["p_memsz"] == next_seg["p_vaddr"]
                    and prev_seg["p_flags"] == next_seg["p_flags"]
                    and prev_seg["p_align"] == next_seg["p_align"]
                ):
                    new_segments.append(
                        Container(
                            **{
                                "p_type": "PT_LOAD",
                                "p_offset": prev_seg["p_offset"],
                                "p_filesz": prev_seg["p_filesz"]
                                + next_seg["p_filesz"]
                                + (
                                    next_seg["p_offset"]
                                    - (prev_seg["p_offset"] + prev_seg["p_filesz"])
                                ),
                                "p_vaddr": prev_seg["p_vaddr"],
                                "p_paddr": prev_seg["p_paddr"],
                                "p_memsz": prev_seg["p_memsz"]
                                + next_seg["p_memsz"]
                                + (
                                    next_seg["p_vaddr"]
                                    - (prev_seg["p_vaddr"] + prev_seg["p_memsz"])
                                ),
                                "p_flags": prev_seg["p_flags"],
                                "p_align": prev_seg["p_align"],
                            }
                        )
                    )
                    i += 2
                else:
                    new_segments.append(prev_seg)
                    i += 1
            if i == len(self._segments) - 1:
                new_segments.append(self._segments[i])
            if new_segments == self._segments:
                break
            self._segments = new_segments

        # generate new phdr at end of the file and update ehdr
        last_seg = sorted(self._segments, key=lambda x: x["p_offset"])[-1]
        phdr_start = last_seg["p_offset"] + last_seg["p_filesz"]
        new_phdr = b""
        for segment in self._segments:
            new_phdr += self._elf.structs.Elf_Phdr.build(segment)
        self.p.binfmt_tool.update_binary_content(phdr_start, new_phdr)

        ehdr = self._elf.header
        ehdr["e_phnum"] = len(self._segments)
        ehdr["e_phoff"] = phdr_start

        # generate new shdr at end of the file and update ehdr
        shdr_start = phdr_start + len(new_phdr)
        new_shdr = b""
        for section in self._sections:
            new_shdr += self._elf.structs.Elf_Shdr.build(section)
        self.p.binfmt_tool.update_binary_content(shdr_start, new_shdr)

        ehdr["e_shnum"] = len(self._sections)
        ehdr["e_shoff"] = shdr_start
        new_ehdr = self._elf.structs.Elf_Ehdr.build(ehdr)
        self.p.binfmt_tool.update_binary_content(0, new_ehdr)


class ElfArmBare(ElfArmLinux):
    @staticmethod
    def detect_target(binary_path):
        return False

    def get_binfmt_tool(self, binfmt_tool, **kwargs):
        binfmt_tool = binfmt_tool or "default"
        if binfmt_tool == "default":
            return CustomElf(self.p, self.binary_path, **kwargs)
        raise NotImplementedError()

    def get_allocation_manager(self, allocation_manager):
        allocation_manager = allocation_manager or "default"
        if allocation_manager == "default":
            return CustomAllocationManager(self.p)
        raise NotImplementedError()



"""
Contains patches that modify the binary at the byte level.
"""

import logging

from .patch import Patch

logger = logging.getLogger(__name__)


class ModifyRawBytesPatch(Patch):
    """
    Patch that modifies bytes of the binary.
    """

    def __init__(self, addr: int, new_bytes: bytes, addr_type="mem") -> None:
        """
        Constructor.

        :param addr: Starting address of bytes you want to change.
        :param new_bytes: New bytes to replace original ones.
        :param addr_type: The type of address given, "mem" (memory address) or "raw" (file address), defaults to "mem"
        """
        self.addr = addr
        self.new_bytes = new_bytes
        self.addr_type = addr_type

    def apply(self, p) -> None:
        """
        Applies the patch to the binary, intended to be called by a Patcherex instance.

        :param p: Patcherex instance.
        :raises NotImplementedError: Raised if an address type other than "raw" or "mem" is specified.
        """
        if self.addr_type == "raw":
            offset = self.addr
        elif self.addr_type == "mem":
            offset = p.binary_analyzer.mem_addr_to_file_offset(self.addr)
            if not offset:
                logger.warning(
                    "failed to convert mem addr to file offset, will just default to raw addr"
                )
                offset = self.addr
        else:
            raise NotImplementedError()
        p.binfmt_tool.update_binary_content(offset, self.new_bytes)



from patcherex2 import *

p = Patcherex("getline")

p.patches.append(InsertInstructionPatch(0x120b,"mov esi, 0xa\n"))
p.patches.append(InsertDataPatch("my_str",b"Ran out of space\0"))

p.patches.append(InsertInstructionPatch(0x1199,"mov [rbp-0xc],esi"))

asm_string = """
    cmp edx, [rbp-0xc]
    jl less
    SAVE_CONTEXT
    lea rdi, [{my_str}]
    call {puts}
    RESTORE_CONTEXT
    mov rsi, [rbp-0x18]
    mov byte ptr [rsi+rax], 0
    mov eax, edx
    leave
    ret
less:
"""

p.patches.append(InsertInstructionPatch(0x11c1,asm_string))

p.apply_patches()

p.save_binary()